import pytest
import pkg_resources
import networkx
import numpy as np
import pandas as pd

from seekr import graph


class TestMaker:

    def _build_adj(self):
        kmers = 'tests/data/example_2mers.npy'
        kmers = pkg_resources.resource_filename('seekr', kmers)
        kmers = np.load(kmers)
        adj = np.corrcoef(kmers) * -1  # Flip signs for fewer negatives
        names = list(range(5))
        adj = pd.DataFrame(adj, names, names)
        return adj

    def _build_disconnected_graph(self):
        g = networkx.Graph()
        edges = [(0, 1), (0, 2), (0, 3), (1, 2), (2, 4), (2, 5), (2, 6), (7, 8), (8, 9)]
        g.add_edges_from(edges)
        networkx.set_edge_attributes(g, 1, 'weight')
        return g

    def _save_adj(self, out_dir, binary=True):
        adj = self._build_adj()
        if binary:
            out_path = str(out_dir.join('adj.npy'))
            np.save(out_path, adj.values)
        else:
            out_path = str(out_dir.join('adj.csv'))
            adj.to_csv(out_path)
        return out_path

    def _get_maker_with_partition(self, out_path):
        gml_path = str(out_path.join('out_main_sub.gml'))
        adj = self._build_adj()
        maker = graph.Maker(adj, gml_path=gml_path, seed=0)
        maker.build()
        maker.save(True)
        maker.get_partition()
        return maker

    def test_apply_threshold(self):
        adj = self._build_adj()
        maker = graph.Maker(adj)
        maker.apply_threshold()
        assert np.alltrue(maker.adj.values.diagonal() == np.zeros(5))
        assert adj.values[1, 0] == 0
        assert adj.values[1, 2] != 0

    def test_apply_threshold_t1(self):
        adj = self._build_adj()
        maker = graph.Maker(adj, threshold=1)
        maker.apply_threshold()
        assert maker.adj.values.sum() == 0

    def test_apply_threshold_ndarray(self):
        adj = self._build_adj().values
        maker = graph.Maker(adj, threshold=1)
        maker.apply_threshold()
        assert maker.adj.sum() == 0

    def test_build(self):
        adj = self._build_adj()
        maker = graph.Maker(adj)
        maker.build()
        assert type(maker.graph) == networkx.Graph
        assert len(maker.graph) == 5
        assert len(maker.graph.edges()) == 9
        assert len(networkx.get_edge_attributes(maker.graph, 'weight')) == 9
        assert maker.adj is None
        assert maker.main_sub is not None

    def test_build_ndarray(self):
        adj = self._build_adj().values
        maker = graph.Maker(adj)
        maker.build()
        assert type(maker.graph) == networkx.Graph
        assert len(maker.graph) == 5
        assert len(maker.graph.edges()) == 9
        assert len(networkx.get_edge_attributes(maker.graph, 'weight')) == 9

    def test_build_no_clear_adj(self):
        adj = self._build_adj()
        maker = graph.Maker(adj)
        maker.build(clear_adj=False)
        assert maker.adj.equals(adj)

    def test_build_no_main_sub(self):
        adj = self._build_adj()
        maker = graph.Maker(adj)
        maker.build(main_sub=False)
        assert maker.main_sub is None

    def test_find_main_sub(self):
        g = self._build_disconnected_graph()
        maker = graph.Maker()
        maker.graph = g
        maker.find_main_sub()
        assert list(maker.main_sub) == list(range(7))

    def test_save(self, tmpdir):
        gml_path = str(tmpdir.join('out.gml'))
        adj = self._build_adj()
        maker = graph.Maker(adj, gml_path=gml_path)
        maker.build()
        maker.save()
        saved = networkx.read_gml(gml_path)
        expected = [str(n) for n in maker.graph.nodes()]
        assert list(saved.nodes()) == expected
        expected = [(str(n1), str(n2)) for n1, n2 in maker.graph.edges()]
        assert list(saved.edges()) == expected

    def test_save_main_sub(self, tmpdir):
        gml_path = str(tmpdir.join('out_main_sub.gml'))
        adj = self._build_adj()
        maker = graph.Maker(adj, gml_path=gml_path)
        maker.build()
        maker.save(main_sub=True)
        saved = networkx.read_gml(gml_path)
        expected = [str(n) for n in maker.main_sub.nodes()]
        assert list(saved.nodes()) == expected
        expected = [(str(n1), str(n2)) for n1, n2 in maker.main_sub.edges()]
        assert list(saved.edges()) == expected

    def test_get_partition(self, tmpdir):
        maker = self._get_maker_with_partition(tmpdir)
        assert np.isclose(maker.partition.modularity, -0.08024691358024699)
        assert maker.partition.membership == [1, 0, 1, 0, 0]

    def test_membership2attribute(self, tmpdir):
        maker = self._get_maker_with_partition(tmpdir)
        name2group = maker.membership2attribute()
        assert name2group == {'0': 1, '1': 0, '2': 1, '3': 0, '4': 0}
        assert name2group == networkx.get_node_attributes(maker.graph, 'Group')

    def test_membership2attribute_disconnected(self, tmpdir):
        gml_path = str(tmpdir.join('out_main_sub.gml'))
        g = self._build_disconnected_graph()
        maker = graph.Maker(gml_path=gml_path)
        maker.graph = g
        maker.find_main_sub()
        maker.save(True)
        maker.get_partition()
        name2group = maker.membership2attribute()
        expected = {0: 1,
                    1: 1,
                    2: 0,
                    3: 1,
                    4: 0,
                    5: 0,
                    6: 0,
                    7: 2,
                    8: 2,
                    9: 2}
        assert name2group == expected

    def test_membership2attribute_disconnected_ncomms1(self, tmpdir):
        gml_path = str(tmpdir.join('out_main_sub.gml'))
        g = self._build_disconnected_graph()
        maker = graph.Maker(gml_path=gml_path, n_comms=1)
        maker.graph = g
        maker.find_main_sub()
        maker.save(True)
        maker.get_partition()
        name2group = maker.membership2attribute()
        expected = {0: 1,
                    1: 1,
                    2: 0,
                    3: 1,
                    4: 0,
                    5: 0,
                    6: 0,
                    7: 1,
                    8: 1,
                    9: 1}
        assert name2group == expected

    def test_membership2attribute_disconnected_ncomms3(self, tmpdir):
        gml_path = str(tmpdir.join('out_main_sub.gml'))
        g = self._build_disconnected_graph()
        maker = graph.Maker(gml_path=gml_path, n_comms=3, gamma=10)
        maker.graph = g
        maker.find_main_sub()
        maker.save(True)
        maker.get_partition()
        name2group = maker.membership2attribute()
        expected = {0: 0,
                    1: 1,
                    2: 2,
                    3: 3,
                    4: 3,
                    5: 3,
                    6: 3,
                    7: 3,
                    8: 3,
                    9: 3}
        assert name2group == expected

    def test_make_gml_file(self, tmpdir):
        gml_path = str(tmpdir.join('out.gml'))
        csv_path = str(tmpdir.join('out.csv'))
        adj = self._build_adj()
        maker = graph.Maker(adj, gml_path=gml_path, csv_path=csv_path)
        maker.make_gml_csv_files()
        in_graph = networkx.read_gml(gml_path)
        assert list(in_graph.nodes()) == [str(i) for i in range(5)]
        assert len(networkx.get_node_attributes(in_graph, 'Group')) == 5
        df = pd.read_csv(csv_path, index_col=0)
        assert np.alltrue(df.index.values == np.arange(5))
        assert np.alltrue(df['Group'].values == np.array([1, 0, 1, 0, 0]))




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

"""

import uuid
import numpy as np
import pandas as pd
import igraph
import networkx
import louvain
import leidenalg

from pathlib import Path

from seekr.utils import get_adj


class Maker:
    """Generate and manipulate network graphs using networkx and igraph.

    Parameters
    ----------
    adj: str | ndarray | DataFrame
        (Path to) adjacency matrix describing weighted edges between nodes.
    gml_path: str
        Path to output file for new network.
    csv_path: str
        Path to two column file storing transcript name and membership.
    leiden: bool (default=True)
        If set to False, use the Louvain algorithm for detection instead of Leiden.
    threshold: float (default=0)
        Value for thresholding adjacency matrix. Below this threshold, all edges are 0.
    gamma: float (default=1)
        Resolution parameter for community detection algorithm.
    n_comms: int (default=5)
        Number of communities to find. This does not count a null community.
    seed: int (default=None)
        Set the seed for reproducible results from community detection.

    Attributes
    ----------
    graph: networkx.Graph
        Primary graph build from adjacency matrix
    main_sub: networkx.Graph
        Largest connected subgraph of primary graph
    partition: RBConfigurationVertexPartition
        Contains the membership data for each node
    df: DataFrame
        Stores transcript name and membership
    tempfile_path: Path
        Location of temporary gml file if gml_path is not specified.
        Needed for converting networkx Graph to igraph Graph.
    """

    def __init__(self, adj=None, gml_path=None, csv_path=None, leiden=True,
                 threshold=0, gamma=1, n_comms=5, seed=None):
        self.adj = adj
        if adj is not None:
            self.adj = get_adj(adj)
        self.gml_path = gml_path
        self.csv_path = csv_path
        self.detector = leidenalg if leiden else louvain
        self.threshold = threshold
        self.gamma = gamma
        self.n_comms = n_comms
        self.seed = seed

        self.graph = None
        self.main_sub = None
        self.partition = None
        self.df = None
        self.tempfile_path = None
        if self.gml_path is None:
            self.tempfile_path = Path(str(uuid.uuid4()) + '.gml')

    def apply_threshold(self):
        """Remove low weighted edges from graph."""
        adj = self.adj
        if isinstance(adj, pd.DataFrame):
            adj = adj.values
        np.fill_diagonal(adj, 0)
        adj[adj < self.threshold] = 0

    def find_main_sub(self):
        """Calculate the largest connected subgraph

        Notes
        -----
        https://en.wikipedia.org/wiki/Connected_component_(graph_theory)
        """
        subgraphs = [self.graph.subgraph(c) for c in networkx.connected_components(self.graph)]
        graph_sizes = [sub.size() for sub in subgraphs]
        self.main_sub = subgraphs[graph_sizes.index(max(graph_sizes))]

    def save(self, main_sub=False):
        """Saves graph in the .gml format

        Parameters
        ----------
        main_sub: bool (default=False)
            If True, the graph stored in self.main_sub will be saved
        """
        graph = self.main_sub if main_sub else self.graph
        # This gives a way to dump the main_sub even when a gml_path hasn't been assigned.
        gml_path = self.gml_path if self.gml_path else self.tempfile_path
        networkx.write_gml(graph, gml_path)

    def get_partition(self, gml_path=None):
        """Run community detection on igraph Graph from .gml file.

        Parameters
        ----------
        gml_path: str (default=None)
            Alternative location for loading in .gml file. Defaults to self.gml_path.
        """
        paths = (gml_path, self.gml_path, self.tempfile_path)
        gml_path = next(path for path in paths if path is not None)
        graph = igraph.Graph.Read_GML(str(gml_path))
        extra_args = {}
        if self.seed is not None:
            if self.detector is louvain:
                self.detector.set_rng_seed(self.seed)
            else:
                extra_args['seed'] = self.seed
        self.partition = self.detector.find_partition(graph,
                                                      self.detector.RBConfigurationVertexPartition,
                                                      weights='weight',
                                                      resolution_parameter=self.gamma,
                                                      **extra_args)

    def membership2attribute(self):
        """Store communities in graph.

        The function maps values from the partition to graph nodes.
        Nodes found in communities with values larger than n_comms are put into the null community.
        Similarly, nodes not found in main_sub will also be added to the null community.

        Returns
        -------
        name2group: {str: int}
            Maps name of transcript to its community in the partition
        """
        n_comms = min(self.n_comms, len(set(self.partition.membership)))
        main2group = dict(zip(self.main_sub.nodes(), self.partition.membership))
        name2group = {}
        for node in self.graph.nodes():
            group = min(n_comms, main2group.get(node, n_comms))
            name2group[node] = group
        networkx.set_node_attributes(self.graph, name='Group', values=name2group)
        return name2group

    def build(self, clear_adj=True, main_sub=True):
        """Makes networkx graph from adjacency matrix

        Parameters
        ----------
        clear_adj: bool (default=True)
            If False, skip running self.adj=None to clear memory.
        main_sub: bool (default=True)
            If False, skip finding main commected subgraph.
        """
        self.apply_threshold()
        if isinstance(self.adj, np.ndarray):
            self.graph = networkx.from_numpy_array(self.adj)
        else:
            self.graph = networkx.from_pandas_adjacency(self.adj)
        networkx.relabel_nodes(self.graph, lambda n: str(n), False)
        if clear_adj:
            self.adj = None
        if main_sub:
            self.find_main_sub()

    def make_gml_csv_files(self):
        """Wrapper function for most common way to generate .gml file."""
        self.build()
        self.save(True)
        self.get_partition()
        name2group = self.membership2attribute()
        if self.gml_path is not None:  # This is outside of the save function intentionally.
            self.save()
        self.df = pd.DataFrame.from_dict(name2group, orient='index')
        self.df.columns = ['Group']
        if self.csv_path is not None:
            self.df.to_csv(self.csv_path)
        if self.tempfile_path:
            self.tempfile_path.unlink()



import numpy as np

from seekr import pearson


class TestPearson:

    def test_pearson(self):
        counts1 = np.array([[8, 5, 6, 9, 2],
                            [8, 3, 6, 6, 7],
                            [7, 7, 3, 3, 7]])
        counts2 = np.array([[ 2,  8, -9, -1, -8],
                            [-4,  1,  2, -1,  2],
                            [ 5, -3, -7,  2, -9]])
        dist = pearson.pearson(counts1, counts2)
        expected = np.array([[0.3217847, -0.71611487, 0.85110363],
                             [-0.52756992, -0.47172818, 0.22652512],
                             [0.43762719, -0.17902872, 0.01547461]])
        assert np.allclose(dist, expected)

    def test_pearson_one(self):
        counts1 = np.array([[1, 2, 3, 4], [2, 4, 6, 8]])
        dist = pearson.pearson(counts1, counts1)
        expected = np.ones((2, 2))
        assert np.allclose(dist, expected)


class TestDomainPearson:
    # TODO (jessime) Add tests!
    pass



import pytest
import pkg_resources
import networkx
import numpy as np
import pandas as pd

from seekr import utils


class TestMaker:

    def _save_adj(self, out_dir, binary=True):
        adj = self._build_adj()
        if binary:
            out_path = str(out_dir.join('adj.npy'))
            np.save(out_path, adj.values)
        else:
            out_path = str(out_dir.join('adj.csv'))
            adj.to_csv(out_path)
        return out_path

    def _build_adj(self):
        kmers = 'tests/data/example_2mers.npy'
        kmers = pkg_resources.resource_filename('seekr', kmers)
        kmers = np.load(kmers)
        adj = np.corrcoef(kmers) * -1  # Flip signs for fewer negatives
        names = list(range(5))
        adj = pd.DataFrame(adj, names, names)
        return adj

    def test_get_adj_ndarray(self):
        expected = self._build_adj().values
        adj = utils.get_adj(expected)
        assert type(adj) == np.ndarray

    def test_get_adj_df(self):
        expected = self._build_adj()
        adj = utils.get_adj(expected)
        assert type(adj) == pd.DataFrame

    def test_get_adj_str_ndarray(self, tmpdir):
        adj_path = self._save_adj(tmpdir)
        adj = utils.get_adj(adj_path)
        assert type(adj) == np.ndarray

    def test_get_adj_str_df(self, tmpdir):
        adj_path = self._save_adj(tmpdir, binary=False)
        adj = utils.get_adj(adj_path)
        assert type(adj) == pd.DataFrame


# -*- coding: utf-8 -*-
"""
Created on Tue Dec 15 11:13:42 2015

@author: jessime

Warnings
--------
* A lot of this code is dependent on GENCODE formatted fasta files.
* The C code in the shuffle function will segfault if the string is too big.
This will cause the Python Kernel to crash.

Issues
------
1. Filters are too similar. Shouldn't repeat that code. Just send in the filter condition.
2. Maker and Extractor share code that maybe should be in parent class
3. Do `if outfile: save \ always return`
"""

import os
import gzip
import pickle
import shutil
import ftplib
import requests
import urllib.request
import numpy as np

from contextlib import closing
from os.path import exists, join
from os import makedirs
from ushuffle import shuffle
from ushuffle import set_seed

from seekr.my_tqdm import my_tqdm, my_trange
from seekr.fasta_reader import Reader


class Maker:
    """Manipulates fasta files into new fasta files.

    Parameters
    ----------
    infasta : str (default=None)
        Name of input fasta file to be manipulated
    outfasta : str (default=None)
        Location to store filtered fasta file
    outnames : str (default=None)
        Location to stores saved transcript names

    Attributes
    ----------
    data : list
        Zipped list of names and seqs tuples
    names : list
        Header lines of fasta file
    seqs : list
        Sequences from fasta file
    """

    def __init__(self, infasta=None, outfasta=None, outnames=None):
        self.infasta = infasta
        if infasta is not None:
            self.data, self.names, self.seqs = Reader(infasta).get_data()
        self.outfasta = outfasta
        self.outnames = outnames

    def _name_dump(self, common_name_list):
        """Save new names if necessary

        Parameters
        ----------
        common_name_list : list
            Common style names of sequences which are being kept
        """
        if self.outnames is not None:
            pickle.dump(common_name_list, open(self.outnames, 'wb'))

    def filter1(self, zeros=1, unique_per_gene=False):
        """Filters a fasta file to only keep *1 transcript names.

        Parameters
        ----------
        zeros: int (default=1)
            The number of zeros preceding `1`.
            (e.g. zeros=1 will filter for `01` transcripts. zeros=2 will filter for `001`)
        unique_per_gene: bool (default=False)
            If True, when multiple 01 isoforms exist per gene, only keep the smaller name.
            (e.g. if XIST-001 and XIST-201 both exist, keep XIST-001)

        Returns
        -------
        common_name_list: List[str]
            The common style names of the sequences which are being kept
        """
        common_name_list = []
        with open(self.outfasta, 'w') as outfasta:
            current_gene = ''
            best_name_for_gene = ''
            best_seq_for_gene = ''
            for name, seq in self.data:
                name_data = name.split('|')
                common_name = name_data[4]
                gene = name_data[1]
                canonical = common_name[-(zeros+1):] == '0'*zeros+'1'
                if canonical:
                    if not unique_per_gene:
                        common_name_list.append(common_name)
                        outfasta.write(name+'\n')
                        outfasta.write(seq+'\n')
                        continue
                    new_gene = gene != current_gene
                    if not best_name_for_gene:
                        best_name_for_gene = name
                        best_seq_for_gene = seq
                        current_gene = gene
                    elif best_seq_for_gene and not new_gene:
                        new_is_better = common_name[-3:] < best_name_for_gene.split('|')[4][-3:]
                        if new_is_better:
                            best_name_for_gene = name
                            best_seq_for_gene = seq
                        warning = (f'Gene {gene} has at least two viable isoforms. '
                                   f'Keeping: {best_name_for_gene}')
                        print(warning)
                    elif new_gene:
                        common_name_list.append(best_name_for_gene.split('|')[4])
                        outfasta.write(best_name_for_gene+'\n')
                        outfasta.write(best_seq_for_gene + '\n')
                        current_gene = gene
                        best_name_for_gene = name
                        best_seq_for_gene = seq
            if best_seq_for_gene:
                common_name_list.append(common_name)
                outfasta.write(best_name_for_gene + '\n')
                outfasta.write(best_seq_for_gene + '\n')
        self._name_dump(common_name_list)
        return common_name_list

    def filter_size(self, size_lim, keep_all_below=True):
        """Filters fasta file based on size of each transcript

        Parameters
        ----------
            size_lim : int
                the size limit to keep above or below
            keep_all_below : bool (default=True)
                Keep sequences longer or shorter than specified length?

        Returns
        -------
        filtered_fasta : list
            Lines of the new fasta file
        """
        common_name_list = []
        filtered_fasta = []
        for name, seq in self.data:
            common_name = name.split('|')[4]
            length = len(seq)
            if ((length <= size_lim and keep_all_below) or
            (length >= size_lim and not keep_all_below)):
                if self.outnames:
                    common_name_list.append(common_name)
                filtered_fasta.append((name, seq))

        if self.outfasta is not None:
            with open(self.outfasta, 'w') as outfasta:
                for dt in filtered_fasta:
                    outfasta.write(dt+'\n')
        self._name_dump(common_name_list)
        return filtered_fasta

    def filter_name(self, keep_names):
        """Filters fasta file based on a list of common names

        Parameters
        ----------
        keep_names : list
            The common style names to be kept from the fasta file

        Returns
        -------
        filtered_fasta : list
            Lines of the new fasta file
        """
        common_names = [n.split('|')[4] for n in self.names]
        names_dict = dict(zip(common_names, self.data))
        filtered_fasta = []
        for n in keep_names:
            try:
                filtered_fasta.append(names_dict[n])
            except KeyError:
                print('{} is not in fasta file'.format(n))

        if self.outfasta is not None:
            with open(self.outfasta, 'w') as outfasta:
                for dt in filtered_fasta:
                    outfasta.write(dt[0]+'\n')
                    outfasta.write(dt[1]+'\n')
        return filtered_fasta

    def separate(self, outdir, filenames=None):
        """Separate a fasta file into individual fasta files

        Parameters
        ----------
        outdir : str
            Path of directory in which to save the new fasta files
        filenames : [str] (default=None)
            Optional list of names for files instead of integers.
        """
        if not exists(outdir):
            makedirs(outdir)

        if filenames is None:
            filenames = range(len(self.data))
        for fn, (name, seq) in zip(filenames, self.data):
            fn = '{}.fa'.format(fn)
            with open(join(outdir, fn), "w") as outfile:
                outfile.write(name+'\n')
                outfile.write(seq+'\n')


class RandomMaker(Maker):
    """Provides functions for creating randomly generated fasta files

    Parameters
    ----------
    infasta: str (default=None)
        Name of input fasta file to be manipulated
    outfasta: str (default=None)
        Location to store filtered fasta file
    k: int (default=1)
        The size of kmer to conserve between original and random sequences
    mutations: int
        Number of SNP mutations to make in sequence
    seed: int (default=None)
        Seed to use for reproducible shuffling with ushuffle
    individual: bool (default=True)
        Whether to conserve kmers of each sequence or the entire fasta file

    Attributes
    ----------
    error: float
        The mean error in kmer frequency betwen original counts and new counts
    weights: DataFrame
        Shows fractions of each kmer for old and new seqs, along with error
    new_fasta_seqs: list
        New, randomized sequences with conserved kmer frequencies

    Notes
    -----
    The kmers are conserved using the uShuffle algorithm described here:
        http://bmcbioinformatics.biomedcentral.com/articles/10.1186/1471-2105-9-192
    The C code can be found here:
        https://github.com/dcjones/cbgb/tree/master/ushuffle
    The Cython code can be found here:
        https://github.com/guma44/ushuffle
    """

    def __init__(self, infasta=None, outfasta=None, k=1, mutations=0, seed=None, individual=True):
        super(RandomMaker, self).__init__(infasta=infasta, outfasta=outfasta)
        self.k = k
        self.mutations = mutations
        self.individual = individual
        self.weights = None
        self.new_fasta_seqs = None
        if seed is not None:
            set_seed(seed)
            np.random.seed(seed)

    def shuffle(self, seq):
        rand_seq = shuffle(seq.encode(), self.k).decode('utf-8')
        rand_array = np.array(list(rand_seq))
        indices = np.random.choice(len(seq), self.mutations, replace=False)
        new_bases = np.random.choice(list('AGTC'), self.mutations)
        rand_array[indices] = new_bases
        rand_seq = ''.join(rand_array)
        return rand_seq

    def get_random_seqs(self, seqs):
        """Generates random RNA sequences

        Parameters
        ----------
        seqs : [str]
            Original sequences to create random copies of

        Returns
        -------
        new_seqs : [str]
            Randomly generated sequences
        """
        new_seqs = [self.shuffle(s) for s in my_tqdm()(seqs)]
        return new_seqs

    def split(self, seq):
        """Return concatenated sequence to individual sequence lengths"""
        lengths = [len(s) for s in self.seqs]
        new_seqs = []
        for l in lengths:
            new_seqs.append(seq[:l])
            seq = seq[l:]
        return new_seqs

    def inject_seqs(self, new_seqs):
        """Use the original fasta file headers to put the random sequences in fasta format

        Parameters
        ----------
        new_seqs : list
            string elements, each a randomly generated RNA

        Returns
        -------
        new_fasta_seqs : list
            The lines of a new fasta file filled with random RNAs
        """
        new_fasta_seqs = []
        for name, new_seq in zip(self.names, new_seqs):
            new_fasta_seqs.append(name)
            new_fasta_seqs.append(new_seq)
        return new_fasta_seqs

    def save(self):
        if self.outfasta is not None:
            with open(self.outfasta, 'w') as outfasta:
                for line in self.new_fasta_seqs:
                    outfasta.write('{}\n'.format(line))

    def synthesize_random(self):
        """Make random RNAs based on natural RNAs.

        Returns
        -------
        new_fasta_seqs : list
            Lines of the new fasta file
        """
        if self.individual:
            new_seqs = self.get_random_seqs(self.seqs)
        else:
            new_seqs = self.get_random_seqs([''.join(self.seqs)])
            new_seqs = self.split(new_seqs[0])
        self.new_fasta_seqs = self.inject_seqs(new_seqs)
        self.save()
        return self.new_fasta_seqs


class Downloader:
    """Download fasta files from a variety of online sources

    #TODO Expand to other sources of fasta files.
    """

    def __init__(self):
        pass

    def find_current_release(self, species):
        """Scrape Genecode's site to find the latest release value.

        Parameters
        ----------
        species: str
            Name of species (human or mouse)
        """
        url = f'https://www.gencodegenes.org/{species}/'
        html = requests.get(url).text
        for line in html.splitlines():
            if '<title>' in line:
                title = line
                break
        release = title.split('Release')[1].strip().strip('</title>')
        return release

    def build_url(self, biotype, species, release):
        """Build the correct ftp URL to download from GENCODE.

        Parameters
        ----------
        biotype: str
            Name of Genocde set to download. Must be one of ('all', 'pc', 'lncRNA').
        species: str (default='human')
            Name of species. Must be one of: ('human' or 'mouse').
        release: str (default=None)
            Name of specific release to download (e.g. 'M5'). If None, download latest release.

        Returns
        -------
        url: str
            FTP file to download.
        release: str
            Name of specific release to download (e.g. 'M5'). Will not be None.
        """
        error_msg = "'biotype' must be in ('all', 'pc', 'lncRNA')."
        assert biotype in ('all', 'pc', 'lncRNA'), error_msg
        error_msg = "'species' must be either 'human' or 'mouse'."
        assert species in ('human', 'mouse'), error_msg
        biotype2prefix = {'all': '', 'pc': 'pc_', 'lncRNA': 'lncRNA_'}
        prefix = biotype2prefix[biotype]
        if release is None:
            release = self.find_current_release(species)
        if species == 'mouse':
            error_msg = "Mouse releases must begin with 'M'."
            assert release[0] == 'M', error_msg
        url_base = 'ftp://ftp.ebi.ac.uk/pub/databases/gencode/Gencode_'
        url = f'{species}/release_{release}/gencode.v{release}.{prefix}transcripts.fa.gz'
        url = url_base + url
        return url, release

    def gunzip(self, gzip_path):
        """Unzip a gzipped file and remove orginal.

        Paramters
        ---------
        gzip_path: str
            Gzipped file location
        """
        out_path = gzip_path.strip('.gz')
        with gzip.open(gzip_path, 'rb') as in_file:
            with open(out_path, 'wb') as out_file:
                shutil.copyfileobj(in_file, out_file)
        os.remove(gzip_path)

    def get_gencode(self, biotype, species='human', release=None, out_path=None, unzip=True):
        """Download .fa.gz file from Gencode's site.

        Parameters
        ----------
        biotype: str
            Name of Genocde set to download. Must be one of ('all', 'pc', 'lncRNA').
        species: str (default='human')
            Name of species. Must be one of: ('human' or 'mouse').
        release: str (default=None)
            Name of specific release to download (e.g. 'M5'). If None, download latest release.
        out_path: str (default=None)
            Path to location for fasta file. Default will save by release name.
        unzip: bool (default=True)
            If False, do not gunzip fasta file after downloading
        """
        url, release = self.build_url(biotype, species, release)
        if out_path is not None:
            error_msg = "Even if unzipping, 'out_path' must end with '.gz'."
            assert out_path.endswith('.gz'), error_msg
        try:
            with closing(urllib.request.urlopen(url)) as r:
                if out_path is None:
                    out_path = f'v{release}_{biotype}.fa.gz'
                with open(out_path, 'wb') as out_file:
                    shutil.copyfileobj(r, out_file)
            if unzip:
                self.gunzip(out_path)
        except urllib.error.URLError as url_error:
            print('The file failed to download because:\n', url_error)
            cd_err = "<urlopen error ftp error: error_perm('550 Failed to change directory.',)>"
            if str(url_error) == cd_err:
                print('Did you pass a valid `--release` value (e.g. M14, 22)?')




# -*- coding: utf-8 -*-
"""
Created on Mon Sep 12 11:16:50 2016

@author: jessime

A couple of functions for autodetecting if code is being run in the notebook.
The appropriate tqdm progressbar will be returned.
"""

from tqdm import tnrange, trange, tqdm, tqdm_notebook


import sys


def _is_kernel():
    if 'IPython' not in sys.modules:
        # IPython hasn't been imported, definitely not
        return False
    from IPython import get_ipython
    # check for `kernel` attribute on the IPython instance
    return getattr(get_ipython(), 'kernel', None) is not None


def my_tqdm():
    return tqdm_notebook if _is_kernel() else tqdm


def my_trange():
    return tnrange if _is_kernel() else trange



# TODO (Dan) What should this file be called?
import pandas as pd
import numpy as np

from collections import defaultdict
from itertools import product
from pathlib import Path


class CountsWeighter:
    """Weight kmer counts by a collection of PWMs.

    Parameters
    ----------
    pwm_dir: str (default=None)
        Path to directory containing pwm files patterned *.txt.
    counts: str | ndarray | DataFrame (default=None)
        (Path to) kmer counts matrix.
    k: int
        Length of kmer.
    out_path: str (default=None)
        Path to .csv file for weighted counts

    Attributes
    ----------
    k_sub: int (default=4)
        Length of sub_kmers to use if motif length is less than k.
        Currently hard-coded to 4.
    kmers: list
        Str elements of all kmers of size k
    df: pd.DataFrame (default=None)
        # TODO (Dan) One line description of contents
    """
    def __init__(self, pwm_dir=None, counts=None, k=5, out_path=None):
        self.pwm_dir = pwm_dir
        if pwm_dir is not None:
            self.pwm_dir = Path(pwm_dir)
        self.counts = counts
        self.k = k
        self.out_path = out_path

        self.k_sub = 4
        self.kmers = [''.join(p) for p in product('AGTC', repeat=self.k)]
        self.df = None
        if counts is not None:  # get_counts depends on self.kmers
            self.counts = self.get_counts(counts)

    def get_counts(self, counts):
        """Load kmer counts matrix from .csv or .npy file, if necessary.

        Parameters
        ----------
        counts: str | ndarray | DataFrame
            (Path to) kmer counts matrix.

        Returns
        -------
        counts: DataFrame
            Kmer counts matrix describing transcript kmer profiles.
        """
        counts_types = (str, pd.DataFrame, np.ndarray)
        err_msg = f'adj must be one of {counts_types}, not {type(counts)}.'
        assert type(counts) in counts_types, err_msg
        if isinstance(counts, str):
            try:
                counts = pd.read_csv(counts, index_col=0)
            except UnicodeDecodeError:
                counts = np.load(counts)
        if isinstance(counts, np.ndarray):
            counts = pd.DataFrame(counts, columns=self.kmers)
        return counts

    def gen_pwm_dicts(self):
        """Search directory for non-empty pwm files, and load them.

        Yields
        ------
        pwm_path: str
            Path to non-empty pwm file.
        pwm: Dict[str: Dict[int: float]]
            Position weight matrix as a nested dict (for fast lookups).
            Outer keys are nucleotides. Inner keys are positions. Inner values are weights.
        """
        for pwm_path in self.pwm_dir.glob('*.txt'):
            try:
                pwm = pd.read_csv(pwm_path, sep='\t')
            except pd.errors.EmptyDataError:
                print(f'The motif file {pwm_path} is empty. Skipping.')
                continue
            pwm.drop('Pos', axis=1, inplace=True)
            pwm = pwm.rename(columns={'U': 'T'}).to_dict()
            yield pwm_path, pwm

    def set_kmer2weight(self, kmer2weight, pwm, key_kmer, iter_kmer, n_kmers):
        for frame in range(n_kmers):
            weight = 1
            for pos, nucleotide in enumerate(iter_kmer):
                weight *= pwm[nucleotide][pos + frame]
            kmer2weight[key_kmer] += weight

    def build_weights_dict(self, pwm):
        """Create a mapping between all kmers their weights in a given PWM.

        Parameters
        ----------
        pwm: Dict[str: Dict[int: float]]
            Position weight matrix.

        Returns
        -------
        kmer2weight: Dict
            Keys are all kmers, values are the kmer's weight in a given pwm.
        """
        kmer2weight = defaultdict(int)
        motif_len = len(pwm['A'])
        if motif_len < self.k:
            n_kmers = motif_len - 4 + 1
            for kmer in self.kmers:
                for sub_kmer in (kmer[i:i+self.k_sub] for i in range(self.k-self.k_sub+1)):
                    self.set_kmer2weight(kmer2weight, pwm, kmer, sub_kmer, n_kmers)
        else:
            for kmer in self.kmers:
                n_kmers = motif_len - self.k + 1
                self.set_kmer2weight(kmer2weight, pwm, kmer, kmer, n_kmers)
        return kmer2weight

    def weight_counts(self, kmer2weight):
        """Weight each kmer in counts matix by corresponding kmer weight found in a PWM.

        Parameters
        ----------
        kmer2weight: Dict
            Keys are all kmers, values are the kmer's weight in a given pwm.

        Returns
        -------
        scores_sums: np.array
            Each element corresponds to sum of weighted kmer profile for given transcript.
            Length of scores_sums is equal to length of counts.
        """
        sorted_weights = np.array([kmer2weight[k] for k in self.counts.columns])
        weighted_z_scores = self.counts.values.copy() * sorted_weights
        scores_sums = weighted_z_scores.sum(axis=1)
        return scores_sums

    def save(self):
        """Save df to csv file."""
        if self.out_path is not None:
            self.df.to_csv(self.out_path, float_format='%.4f')

    def run(self):
        """TODO (Dan) Update based on description of self.df"""
        score_dict = {}
        for pwm_path, pwm in self.gen_pwm_dicts():
            kmer2weight = self.build_weights_dict(pwm)
            score_dict[pwm_path.name] = self.weight_counts(kmer2weight)
        self.df = pd.DataFrame.from_dict(score_dict, orient='index', columns=self.counts.index)
        self.save()



import sys
import requests
import gzip
import pkg_resources

from pathlib import Path

import pytest

from seekr import fasta


class TestMaker:

    def test_filter01(self, tmpdir):
        infasta = 'tests/data/v22_pc_head.fa'
        infasta = pkg_resources.resource_filename('seekr', infasta)
        maker = fasta.Maker(infasta, str(Path(tmpdir, 'out.fa')))
        names = maker.filter1()
        assert names == ['OR4F5-001', 'FO538757.3-201', 'FO538757.2-201']

    def test_filter001(self, tmpdir):
        infasta = 'tests/data/v22_pc_head.fa'
        infasta = pkg_resources.resource_filename('seekr', infasta)
        maker = fasta.Maker(infasta, str(Path(tmpdir, 'out.fa')))
        names = maker.filter1(2)
        assert names == ['OR4F5-001']

    def test_filter01_1per(self, tmpdir, capsys):
        infasta = 'tests/data/example2.fa'
        infasta = pkg_resources.resource_filename('seekr', infasta)
        maker = fasta.Maker(infasta, str(Path(tmpdir, 'out.fa')))
        names = maker.filter1(unique_per_gene=True)
        assert names == ['JK-001', 'JK2-001']
        captured = capsys.readouterr()
        expected = ('Gene ENSG1 has at least two viable isoforms. '
                    'Keeping: >ENST1|ENSG1|OTTHUMG1|OTTHUMT1|JK-001|JK|918|CDS:1-918|\n'
                    'Gene ENSG2 has at least two viable isoforms. '
                    'Keeping: >ENST4|ENSG2|OTTHUMG2|OTTHUMT4|JK2-001|JK2|918|CDS:1-918|\n')
        assert captured.out == expected


class TestRandomMaker:

    @pytest.mark.skipif(sys.platform == 'darwin', reason='Random seed fails on Mac')
    def test_shuffle1(self):
        rand_maker = fasta.RandomMaker(seed=1)
        rand_seq = rand_maker.shuffle('AGTCAGTC')
        assert rand_seq == 'TTGGACAC'

    def test_shuffle2(self):
        rand_maker = fasta.RandomMaker(k=2, seed=1)
        rand_seq = rand_maker.shuffle('AGTCAGTCAGTCAGTC')
        assert rand_seq == 'AGTCAGTCAGTCAGTC'
        kmer_counts = {
            'AA': 0,
            'AG': 4,
            'AT': 0,
            'AC': 0,
            'GA': 0,
            'GG': 0,
            'GT': 4,
            'GC': 0,
            'TA': 0,
            'TG': 0,
            'TT': 0,
            'TC': 4,
            'CA': 3,
            'CG': 0,
            'CT': 0,
            'CC': 0
        }
        rand_kmer_counts = {k:rand_seq.count(k) for k in kmer_counts}
        assert kmer_counts == rand_kmer_counts

    def test_mutations(self):
        rand_maker = fasta.RandomMaker(seed=1)
        rand_maker.mutations = 2
        rand_seq = rand_maker.shuffle('AAAAAA')
        assert rand_seq == 'ACGAAA'

    @pytest.mark.skipif(sys.platform == 'darwin', reason='Random seed fails on Mac')
    def test_get_random_seqs(self):
        rand_maker = fasta.RandomMaker(k=2, seed=1)
        seqs = [
            'AGTCAGTCAGTCAGTC',
            'ATGATATATATATGAT',
            'ATGCATAGTTTTTTTTTCTGC'
        ]
        rand_seqs = rand_maker.get_random_seqs(seqs)
        expected = ['AGTCAGTCAGTCAGTC',
                    'ATGATGATATATATAT',
                    'ATTGCTTTGTTTAGCATTTTC']
        assert rand_seqs == expected

    def test_split(self):
        rand_maker = fasta.RandomMaker(k=2, seed=1)
        seqs = [
            'this sentence is 35 characters long',
            'this one is 14'
        ]
        rand_maker.seqs = seqs
        seq = 'TCATTAAGCGCGTCGGTCTCTGTGTACGTCATCTCCATTTTTTTTCGTG'
        rand_seqs = rand_maker.split(seq)
        expected = ['TCATTAAGCGCGTCGGTCTCTGTGTACGTCATCTC',
                    'CATTTTTTTTCGTG']
        assert rand_seqs == expected
        assert len(expected[0]) == 35
        assert len(expected[1]) == 14

    def test_inject_seqs(self):
        rand_maker = fasta.RandomMaker()
        rand_maker.names = ['>seq1', '>seq2']
        new_seqs = ['this is new', 'also new']
        new_fasta_seqs = rand_maker.inject_seqs(new_seqs)
        expected = ['>seq1', 'this is new', '>seq2', 'also new']
        assert new_fasta_seqs == expected

    @pytest.mark.skipif(sys.platform == 'darwin', reason='Random seed fails on Mac')
    def test_synthesize_random(self, tmpdir):
        infasta = 'tests/data/example.fa'
        infasta = pkg_resources.resource_filename('seekr', infasta)
        outfasta = str(tmpdir.join('rand.fa'))
        rand_maker = fasta.RandomMaker(infasta, outfasta, seed=1)
        rand_maker.synthesize_random()
        with open(outfasta) as outfasta:
            rand_fasta = ''.join(next(outfasta) for i in range(6))
        expected = ('>SEQ1\n'
                    'AAAAAA\n'
                    '>SEQ2\n'
                    'GTGTGTGTGTTG\n'
                    '>SEQ3\n'
                    'ATTACTTGGCACCGGA\n')
        assert rand_fasta == expected


class TestDownloader:

    def _expected_current_release(self, species):
        # Note, this information will change over time, so you can't hard code a value.
        # My strategy is to get the value a second way and ensure it's the same.
        url = f'https://www.gencodegenes.org/{species}/'
        html = requests.get(url).text
        for line in html.splitlines():
            if line.startswith('<h1>'):
                header = line
                break
        release = header.split()[1]
        return release

    def test_find_current_release(self):
        release = fasta.Downloader().find_current_release('human')
        expected = self._expected_current_release('human')
        assert expected == release

    def test_gunzip(self, tmpdir):
        test_path = Path(tmpdir, 'test.txt.gz')
        unzipped = Path(tmpdir, 'test.txt')
        content = 'Hello, World!'.encode()
        with gzip.open(test_path, 'wb') as test_file:
            test_file.write(content)
        fasta.Downloader().gunzip(str(test_path))
        assert not test_path.exists()
        assert unzipped.exists()
        assert unzipped.read_text() == content.decode()

    def test_build_url(self):
        downloader = fasta.Downloader()
        url, release = downloader.build_url('lncRNA', 'human', None)
        exp_release = self._expected_current_release('human')
        expected = ('ftp://ftp.ebi.ac.uk/pub/databases/gencode/Gencode_'
                    f'human/release_{exp_release}/gencode.v{exp_release}.lncRNA_transcripts.fa.gz')
        assert url == expected

    def test_build_url_mouse(self):
        downloader = fasta.Downloader()
        url, release = downloader.build_url('all', 'mouse', None)
        exp_release = self._expected_current_release('mouse')
        expected = ('ftp://ftp.ebi.ac.uk/pub/databases/gencode/Gencode_'
                    f'mouse/release_{exp_release}/gencode.v{exp_release}.transcripts.fa.gz')
        assert url == expected

    def test_build_url_pc22(self):
        downloader = fasta.Downloader()
        url, release = downloader.build_url('pc', 'human', '22')
        expected = ('ftp://ftp.ebi.ac.uk/pub/databases/gencode/Gencode_'
                    'human/release_22/gencode.v22.pc_transcripts.fa.gz')
        assert url == expected

    # TODO Consider using mocking to help with testing downloads, or move to integration tests...
    # def test_get_gencode(self, tmpdir):
    #     downloader = fasta.Downloader()
    #     out_path = Path(tmpdir, 'human_lncs.fa.gz')
    #     unzipped = Path(tmpdir, 'human_lncs.fa')
    #     downloader.get_gencode('lncRNA', out_path=str(out_path))
    #     assert not out_path.exists()
    #     assert unzipped.exists()
    #     with unzipped.open() as in_file:
    #         line = next(in_file)
    #         assert line.startswith('>')
    #
    # def test_get_gencode_mouse(self, tmpdir):
    #     downloader = fasta.Downloader()
    #     out_path = Path(tmpdir, 'mouse_all.fa.gz')
    #     unzipped = Path(tmpdir, 'mouse_all.fa')
    #     downloader.get_gencode('all', 'mouse', out_path=str(out_path))
    #     assert not out_path.exists()
    #     assert unzipped.exists()
    #     with unzipped.open() as in_file:
    #         line = next(in_file)
    #         assert line.startswith('>')
    #
    # def test_get_gencode_pc22(self, tmpdir):
    #     downloader = fasta.Downloader()
    #     out_path = Path(tmpdir, 'human_pc_22.fa.gz')
    #     unzipped = Path(tmpdir, 'human_pc_22.fa')
    #     downloader.get_gencode('pc', release='22', out_path=str(out_path))
    #     assert not out_path.exists()
    #     assert unzipped.exists()
    #     with unzipped.open() as in_file:
    #         # Since this file is both "locked" and unzipped, let's check the line count.
    #         count = len(in_file.readlines())
    #         assert count == 187052
    #
    # def test_get_gencode_zipped(self, tmpdir):
    #     downloader = fasta.Downloader()
    #     out_path = Path(tmpdir, 'mouse_lncs.fa.gz')
    #     downloader.get_gencode('lncRNA', 'mouse', 'M5', out_path=str(out_path), unzip=False)
    #     assert out_path.exists()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Description
-----------
Generates a kmer count matrix of m rows by n columns,
where m is the number of transcripts in a fasta file and n is 4^kmer.

Examples
--------
Generate a small, plain text .csv file:

```
from seekr.kmer_counts import BasicCounter

counter = BasicCounter(infasta='example.fa',
                       outfile='example_2mers.csv',
                       k=2,
                       binary=False)
counts = counter.make_count_file()
```

Notes
-----


Issues
------
Any issues can be reported to https://github.com/CalabreseLab

---
"""

import numpy as np

from enum import Enum
from collections import defaultdict
from itertools import product
from pandas import DataFrame

from seekr.my_tqdm import my_tqdm
from seekr.fasta_reader import Reader


Log2 = Enum('Log2', ('pre', 'post', 'none'))


class BasicCounter:
    """Generates overlapping kmer counts for a fasta file

    Parameters
    ----------
    infasta: str (default=None)
        Full path to fasta file to be counted
    outfile: str (default=None)
        Full path to the counts file to be saved
    k: int (default=6)
        Size of kmer to be counted
    binary: bool (default=True)
        Saves as numpy array if True, else saves as csv
    mean: bool, np.array, str (default=True)
        Set the mean to 0 for each kmer/column of the count matrix.
        If str, provide path to a previously calculated mean array.
    std: bool, np.array, str (default=True)
        Set the std. dev. to 1 for each kmer/column of the count matrix.
        If str, provide path to a previously calculated std array.
    log2: Log2 (default=Log2.post)
        Log2 transformation can occur pre- or post-standardization, or not at all.
    leave: bool (default=True)
        Set to False if get_counts is used within another tqdm loop
    silent: bool (default=False)
        Set to True to turn off tqdm progress bar
    alphabet: str (default='AGTC')
        Valid letters to include in kmer.

    Attributes
    ----------
    counts: np.array
        Matrix of kmer counts. Dimensions are equal to # of transcripts by # of kmers.
    kmers: list
        Str elements of all kmers of size k
    map: dict
        Mapping of kmers to column values
    alpha_len: int
        Length of alphabet
    """

    def __init__(self, infasta=None, outfile=None, k=6,
                 binary=True, mean=True, std=True, log2=Log2.post,
                 leave=True, silent=False, label=False, alphabet='AGTC'):
        self.infasta = infasta
        self.seqs = None
        if infasta is not None:
            self.seqs = Reader(infasta).get_seqs()
        self.outfile = outfile
        self.k = k
        self.binary = binary
        self.mean = mean
        if isinstance(mean, str):
            self.mean = np.load(mean)
        self.std = std
        if isinstance(std, str):
            self.std = np.load(std)
        self.log2 = log2 
        self.leave = leave
        self.silent = silent
        self.label = label
        self.counts = None
        self.alpha_len = len(alphabet)
        self.kmers = [''.join(i) for i in product(alphabet, repeat=k)]
        self.map = {k: i for k, i in zip(self.kmers, range(self.alpha_len**k))}

        if self.seqs is not None:
            if len(self.seqs) == 1 and self.std is True:
                err = ('You cannot standardize a single sequence. '
                       'Please pass the path to an std. dev. array, '
                       'or use raw counts by setting std=False.')
                raise ValueError(err)

        if not isinstance(self.log2, Log2):
            raise TypeError(f'log2 must be one of {list(Log2)}')

    def occurrences(self, row, seq):
        """Counts kmers on a per kilobase scale"""
        counts = defaultdict(int)
        length = len(seq)
        increment = 1000/(length-self.k+1) #fix, original 1000/length, incorrect divisor, should be 1000/(l-k+1)
        for c in range(length-self.k+1):
            kmer = seq[c:c+self.k]
            counts[kmer] += increment
        for kmer, n in counts.items():
            if kmer in self.map:
                row[self.map[kmer]] = n
        return row

    def _progress(self):
        """Determine which iterator to loop over for counting"""
        if self.silent:
            return self.seqs

        if not self.leave:
            tqdm_seqs = my_tqdm()(self.seqs, desc='Kmers', leave=False)
        else:
            tqdm_seqs = my_tqdm()(self.seqs)

        return tqdm_seqs

    def center(self):
        """Mean center counts by column"""
        if self.mean is True:
            self.mean  = np.mean(self.counts, axis=0)
        self.counts -= self.mean

    def standardize(self):
        """Divide out the standard deviations from columns of the count matrix"""
        if self.std is True:
            self.std = np.std(self.counts, axis=0)
        self.counts /= self.std
        if np.isnan(self.counts).any():
            print(('\nWARNING: You have `np.nan` values in your counts '
                   'after standardization. This is likely due to '
                   'a kmer not appearing in any of your sequences. '
                   'Try: \n1) using a smaller kmer size, \n2) beginning '
                   'with a larger set of sequences, \n3) passing '
                   'precomputed normalization vectors from a larger '
                   'data set (e.g. GENCODE).'))

    def log2_norm(self):
        """Apply a log2 transform to the count matrix"""
        self.counts += 1
        self.counts = np.log2(self.counts)

    def get_counts(self):
        """Generates kmer counts for a fasta file"""
        self.counts = np.zeros([len(self.seqs), self.alpha_len**self.k], dtype=np.float32)
        seqs = self._progress()

        for i, seq in enumerate(seqs):
            self.counts[i] = self.occurrences(self.counts[i], seq)
        if self.log2 == Log2.pre:
            self.log2_norm()
        if self.mean is not False:
            self.center()
        if self.std is not False:
            self.standardize()
        if self.log2 == Log2.post:
            self.counts += np.abs(np.min(self.counts))
            self.log2_norm()

    def save(self, names=None):
        """Saves the counts appropriately based on current settings.

        There are four output methods for the counts:
        1. Binary. This saves just the counts as a binary numpy array.
        2. No labels. Saves in plain text, but without any labels.
        3. Default names. If no names are provided, fasta headers will be used as labels.
        4. Custom names. Provide a list of names if you want to label lncRNAs with your own names.

        Parameters
        ----------
        names : [str] (default=None)
            Unique names for rows of the Dataframe.
        """
        err_msg = ('You cannot label a binary file. '
                   'Set only one of "binary" or "label" as True. '
                   'If you used `-b` from the command line, '
                   'try also using `-rl`.')
        assert not (self.binary and self.label), err_msg
        assert self.outfile is not None, 'Please provide an outfile location.'
        if self.binary:
            np.save(self.outfile, self.counts)
        elif self.label:
            if names is None:
                names = Reader(self.infasta).get_headers()
            df = DataFrame(data=self.counts, index=names, columns=self.kmers)
            df.to_csv(self.outfile)
        else:
            np.savetxt(self.outfile, self.counts, delimiter=',', fmt='%1.6f')

    def make_count_file(self, names=None):
        """Wrapper function for the most common way to generate count files.

        Given a numpy file name, it will save a numpy file where counts have been:
        cast as a dense array, centered, and standardized.

        Parameters
        ----------
        names : [str] (default=None)
            lncRNA names to pass to self.save

        Returns
        -------
        counts: np.array
            Matrix of kmer counts. Dimensions are equal to # of transcripts by # of kmers.
        """
        self.get_counts()
        if self.outfile is not None:
            self.save(names)
        return self.counts


"""These tests are treated a little like integration tests."""

import sys
import networkx
import numpy as np
import pandas as pd
import pkg_resources


import pytest

from seekr import console_scripts
from seekr.kmer_counts import Log2
from seekr import graph

class TestConsoleScripts:

    # TODO Move to integration testing or somewhere that doesn't mess with Travis
    # def test_run_download_gencode(self, tmpdir):
    #     out_path = Path(tmpdir, 'lncs.fa.gz')
    #     unzipped = Path(tmpdir, 'lncs.fa')
    #     console_scripts._run_download_gencode(biotype='lncRNA',
    #                                           species='human',
    #                                           release='20',
    #                                           out_path=str(out_path),
    #                                           unzip=True)
    #     assert not out_path.exists()
    #     assert unzipped.exists()
    #     with unzipped.open() as in_file:
    #         count = len(in_file.readlines())
    #         assert count == 48978

    def test_run_kmer_counts(self, tmpdir):
        infasta = 'tests/data/example.fa'
        infasta = pkg_resources.resource_filename('seekr', infasta)
        outfile = str(tmpdir.join('2mers.npy'))
        console_scripts._run_kmer_counts(fasta=infasta,
                                         outfile=outfile,
                                         kmer=2,
                                         binary=True,
                                         centered=True,
                                         standardized=True,
                                         log2=Log2.post.name,
                                         remove_labels=True,
                                         mean_vector=None,
                                         std_vector=None,
                                         alphabet='AGTC')
        kmers = np.load(outfile)
        expected = 'tests/data/example_2mers_counts.npy'
        expected = pkg_resources.resource_filename('seekr', expected)
        expected = np.load(expected)
        assert np.allclose(kmers, expected)

    def test_run_kmer_counts_raw_csv(self, tmpdir):
        infasta = 'tests/data/example.fa'
        infasta = pkg_resources.resource_filename('seekr', infasta)
        outfile = str(tmpdir.join('3mers.csv'))
        console_scripts._run_kmer_counts(fasta=infasta,
                                         outfile=outfile,
                                         kmer=3,
                                         binary=False,
                                         centered=False,
                                         standardized=False,
                                         log2=Log2.none.name,
                                         remove_labels=True,
                                         mean_vector=None,
                                         std_vector=None,
                                         alphabet='AGTC')
        kmers = pd.read_csv(outfile, header=None)
        expected = 'tests/data/example_3mers_raw.csv'
        expected = pkg_resources.resource_filename('seekr', expected)
        expected = pd.read_csv(expected, header=None)
        print(kmers.values)
        print(expected)
        assert np.allclose(kmers.values, expected.values)
    def test_run_kmer_counts_vectors(self, tmpdir):
        infasta = 'tests/data/example.fa'
        infasta = pkg_resources.resource_filename('seekr', infasta)
        mean_vector = 'tests/data/example_mean.npy'
        mean_vector = pkg_resources.resource_filename('seekr', mean_vector)
        std_vector = 'tests/data/example_std.npy'
        std_vector = pkg_resources.resource_filename('seekr', std_vector)
        outfile = str(tmpdir.join('2mers_vectors.npy'))
        console_scripts._run_kmer_counts(fasta=infasta,
                                         outfile=outfile,
                                         kmer=2,
                                         binary=True,
                                         centered=False,
                                         standardized=False,
                                         log2=Log2.post.name,
                                         remove_labels=True,
                                         mean_vector=mean_vector,
                                         std_vector=std_vector,
                                         alphabet='AGTC')
        kmers = np.load(outfile)
        expected = 'tests/data/example_2mers_count.npy'
        expected = pkg_resources.resource_filename('seekr', expected)
        expected = np.load(expected)
        assert np.allclose(kmers, expected)

    def test_run_norm_vectors(self, tmpdir):
        infasta = 'tests/data/example.fa'
        infasta = pkg_resources.resource_filename('seekr', infasta)
        mean = str(tmpdir.join('mean.npy'))
        std = str(tmpdir.join('std.npy'))
        console_scripts._run_norm_vectors(fasta=infasta,
                                          mean_vector=mean,
                                          std_vector=std,
                                          log2=Log2.none.name,
                                          kmer=2)
        mean = np.load(mean)
        std = np.load(std)
        expected_mean = 'tests/data/example_mean.npy'
        expected_mean = pkg_resources.resource_filename('seekr', expected_mean)
        expected_mean = np.load(expected_mean)
        expected_std = 'tests/data/example_std.npy'
        expected_std = pkg_resources.resource_filename('seekr', expected_std)
        expected_std = np.load(expected_std)
        assert np.allclose(mean, expected_mean)
        assert np.allclose(std, expected_std)

    def test_run_graph(self, tmpdir):
        kmers = 'tests/data/example_2mers.npy'
        kmers = pkg_resources.resource_filename('seekr', kmers)
        kmers = np.load(kmers)
        adj = np.corrcoef(kmers) * -1  # Flip signs for fewer negatives
        names = list(range(5))
        adj = pd.DataFrame(adj, names, names)
        adj_path = str(tmpdir.join('adj.csv'))
        adj.to_csv(adj_path)
        gml_path = str(tmpdir.join('graph.gml'))
        csv_path = str(tmpdir.join('communities.csv'))
        console_scripts._run_graph(adj=adj,
                                   gml_path=gml_path,
                                   csv_path=csv_path,
                                   louvain=False,
                                   threshold=.15,
                                   resolution=1,
                                   n_comms=5,
                                   seed=0)
        in_graph = networkx.read_gml(gml_path)
        in_df = pd.read_csv(csv_path, index_col=0)
        assert list(in_graph.nodes()) == [str(i) for i in range(5)]
        expected_weight = in_graph.edges()[('0', '2')]['weight']
        assert np.isclose(expected_weight, 0.5278957407763157)
        assert np.alltrue(in_df['Group'].values == np.array([0, 1, 0, 0, 1]))

    @pytest.mark.skipif(sys.platform == 'darwin', reason='Random seed fails on Mac')
    def test_run_gen_rand_rnas(self, tmpdir):
        infasta = 'tests/data/example2.fa'
        infasta = pkg_resources.resource_filename('seekr', infasta)
        outfasta = str(tmpdir.join('rand.fa'))
        k = '2'
        mutations = '5'
        seed = '0'
        group = False
        console_scripts._run_gen_rand_rnas(infasta, outfasta, k, mutations, seed, group)
        with open(outfasta) as outfasta:
            rand_fasta = ''.join(next(outfasta) for i in range(6))
        expected = ('>ENST1|ENSG1|OTTHUMG1|OTTHUMT1|JK-001|JK|918|CDS:1-918|\n'
                    'TATCAG\n'
                    '>ENST2|ENSG1|OTTHUMG1|OTTHUMT2|JK-201|JK|918|CDS:1-918|\n'
                    'ATACGCA\n'
                    '>ENST3|ENSG2|OTTHUMG2|OTTHUMT3|JK2-201|JK2|918|CDS:1-918|\n'
                    'CTACGCT\n')
        assert rand_fasta == expected




import numpy as np
import pandas as pd


def get_adj(adj):
    """Load adjacency matrix from .csv or .npy file, if necessary.

    Parameters
    ----------
    adj: str | ndarray | DataFrame
        (Path to) adjacency matrix describing weighted edges between nodes.

    Returns
    -------
    adj: ndarray | DataFrame
        Adjacency matrix describing weighted edges between nodes.
    """
    adj_types = (str, pd.DataFrame, np.ndarray)
    err_msg = f'adj must be one of {adj_types}, not {type(adj)}.'
    assert type(adj) in adj_types, err_msg
    if isinstance(adj, str):
        try:
            adj = pd.read_csv(adj, index_col=0)
        except UnicodeDecodeError:
            adj = np.load(adj)
    return adj



# -*- coding: utf-8 -*-
"""
Created on Mon Jul  4 13:10:49 2016

@author: jessime
"""


class Reader():
    """Fixes any compatibility issues a fasta file might have with this code.

    Parameters
    ----------
    infasta : str (default=None)
        Name of input fasta file to be manipulated
    outfasta : str (default=None)
        location to store extracted data from infasta
    names : iter (default=None)
        Common style names to use in header lines

    Attributes
    ----------
    data : list
        Raw lines of the infasta file
        Note: This is different than the data attribute in other classes

    Examples
    --------
    Putting the sequence on one line instead of breaking it every 80 chars.
    Making sure the whole sequence is capitalized.
    Restructuring the name line to work with GENCODE's naming.
    """

    def __init__(self, infasta=None, outfasta=None, names=None):
        self.infasta = infasta
        self.outfasta = outfasta
        self.names = names

        self.data = None

    def _read_data(self):
        """Sets data to stripped lines from the fasta file
        """
        with open(self.infasta) as infasta:
            self.data = [l.strip() for l in infasta]

    def _upper_seq_per_line(self):
        """Sets data to upper case, single line sequences for each header
        """
        new_data = []
        seq = ''
        for i, line in enumerate(self.data):
            if line[0] == '>':
                if seq:
                    new_data.append(seq.upper())
                    seq = ''
                else:
                    assert i == 0, 'There may be a header without a sequence at line {}.'.format(i)
                new_data.append(line)
            else:
                seq += line
        new_data.append(seq.upper())
        self.data = new_data

    def get_lines(self):
        self._read_data()
        self._upper_seq_per_line()
        return self.data

    def get_seqs(self):
        clean_data = self.get_lines()
        seqs = clean_data[1::2]
        return seqs

    def get_headers(self):
        clean_data = self.get_lines()
        headers = clean_data[::2]
        return headers

    def get_data(self, tuples_only=False):
        clean_data = self.get_lines()
        headers = clean_data[::2]
        seqs = clean_data[1::2]
        tuples = zip(headers, seqs)
        if tuples_only:
            return tuples
        else:
            return tuples, headers, seqs

    def supply_basic_header(self):
        """Convert headerlines to GENCODE format with only common name and length"""
        new_fasta = []
        
        if self.names is None:
            self.names = iter(self.get_headers())
        for i, line in enumerate(self.data):
            if line[0] =='>':
                name = next(self.names).strip('>')
                length = len(self.data[i+1])
                new_fasta.append('>||||{}||{}|'.format(name, length))
            else:
                new_fasta.append(line)
        return new_fasta

    def save(self):
        """Write self.data to a new fasta file"""
        with open(self.outfasta, 'w') as outfasta:
            for line in self.data:
                outfasta.write(line+'\n')



# TODO (Dan) rename file to match pwm.py
import numpy as np
import pandas as pd
import pkg_resources

from pathlib import Path
from itertools import product
from collections import defaultdict

from seekr.pwm import CountsWeighter


class TestCountsWeighter:

    def minimal_pwm(self):
        pwm = {'A': {0: .5, 1: .5, 2: .95},
               'G': {0: .1, 1: .2, 2: .05},
               'T': {0: .3, 1: 0., 2: 0},
               'C': {0: .1, 1: .3, 2: 0}}
        return pwm

    def test_get_adj_ndarray(self):
        cw = CountsWeighter(k=1)
        array = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
        counts = cw.get_counts(array)
        expected = pd.DataFrame(array, columns=['A', 'G', 'T', 'C'])
        assert expected.equals(counts)

    def test_get_adj_df(self):
        cw = CountsWeighter(k=1)
        array = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
        expected = pd.DataFrame(array, columns=['A', 'G', 'T', 'C'])
        counts = cw.get_counts(expected)
        assert expected.equals(counts)

    def test_get_adj_str_ndarray(self, tmpdir):
        cw = CountsWeighter(k=1)
        array = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
        array_path = str(Path(tmpdir, 'out.npy'))
        np.save(array_path, array)
        counts = cw.get_counts(array_path)
        expected = pd.DataFrame(array, columns=['A', 'G', 'T', 'C'])
        assert expected.equals(counts)

    def test_get_pwm_dicts(self):
        pwm_dir = pkg_resources.resource_filename('seekr', 'tests/data/pwms/')
        cw = CountsWeighter(pwm_dir, k=1)
        pwm_path, pwm = next(cw.gen_pwm_dicts())
        expected = Path(pkg_resources.resource_filename('seekr', 'tests/data/pwms/M001_0.6.txt'))
        assert pwm_path == expected
        assert len(pwm) == 4
        assert len(pwm['A']) == 7
        assert pwm['A'][0] == 0.39532879396435

    def test_set_kmer2weight(self):
        kmer2weight = defaultdict(int)
        pwm = self.minimal_pwm()
        cw = CountsWeighter(k=2)
        for kmer in (''.join(p) for p in product('AGTC', repeat=2)):
            cw.set_kmer2weight(kmer2weight, pwm, kmer, kmer, 2)
        assert kmer2weight['AA'] == (.5 * .5) + (.5 * .95)
        assert kmer2weight['GG'] == (.1 * .2) + (.2 * .05)
        assert kmer2weight['CC'] == (.1 * .3) + (.3 * 0)
        assert kmer2weight['AG'] == (.5 * .2) + (.5 * 0.05)
        assert len(kmer2weight) == 16

    def test_build_weights_dict_minimal(self):
        pwm = self.minimal_pwm()
        cw = CountsWeighter(k=2)
        kmer2weight = cw.build_weights_dict(pwm)
        assert kmer2weight['AA'] == (.5 * .5) + (.5 * .95)
        assert kmer2weight['GG'] == (.1 * .2) + (.2 * .05)
        assert kmer2weight['CC'] == (.1 * .3) + (.3 * 0)
        assert kmer2weight['AG'] == (.5 * .2) + (.5 * 0.05)
        assert len(kmer2weight) == 16

    def test_build_weights_dict_full(self):
        pwm = pkg_resources.resource_filename('seekr', 'tests/data/pwms/M001_0.6.txt')
        pwm = pd.read_csv(pwm, sep='\t').rename(columns={'U': 'T'}).to_dict()
        cw = CountsWeighter(k=2)
        kmer2weight = cw.build_weights_dict(pwm)
        assert kmer2weight['AA'] == 0.9749391864711447
        assert kmer2weight['CG'] == 0.00473602191097646
        assert kmer2weight['CT'] == 0.1186835711075973
        assert kmer2weight['AG'] == 0.02190317370024123
        assert len(kmer2weight) == 16

    def test_weight_counts(self):
        kmers = 'AGTC'
        kmer2weight = dict(zip(kmers, range(4)))
        cw = CountsWeighter(k=1)
        counts = [[1, 2, 3, 4],
                  [1, 1, 1, 1],
                  [1, 2, 1, 2]]
        counts = pd.DataFrame(np.array(counts), columns=list(kmers))
        cw.counts = counts
        sum_scores = cw.weight_counts(kmer2weight)
        assert np.all(sum_scores == np.array([20,  6, 10]))

    def test_run(self):
        pwm_dir = pkg_resources.resource_filename('seekr', 'tests/data/pwms/')
        counts = pkg_resources.resource_filename('seekr', 'tests/data/example_2mers.npy')
        cw = CountsWeighter(pwm_dir, counts, k=2)
        cw.run()
        expected = np.array([[4.56194883, 5.65991574, 6.5880013, 6.47169944, 8.14747599]])
        assert np.allclose(cw.df.values, expected)



import os

from setuptools import setup


HERE = os.path.abspath(os.path.dirname(__file__))
about = {}
version = os.path.join(HERE, 'seekr', '__version__.py')
with open(version, 'r', encoding='utf-8') as f:
    exec(f.read(), about)

with open("README.md", "r") as fh:
    long_description = fh.read()

requirements = [
    'cython',
    'tqdm',
    'numpy',
    'pandas',
    'requests',
    'networkx',
    'python-igraph',
    'louvain',
    'leidenalg',
    'ushuffle',
    'matplotlib',
    'seaborn'
]

test_requirements = [
    'pytest'
]

setup(name=about['__title__'],
      python_requires='>3.6',
      version=about['__version__'],
      install_requires=requirements,
      tests_require=test_requirements,
      description=about['__description__'],
      long_description=long_description,
      long_description_content_type="text/markdown",
      url=about['__url__'],
      author=about['__author__'],
      author_email=about['__author_email__'],
      license=about['__license__'],
      packages=['seekr'],
      zip_safe=False,
      classifiers=['Intended Audience :: Science/Research',
                   'License :: OSI Approved :: MIT License',
                   'Natural Language :: English',
                   'Programming Language :: Python :: 3.6'],
      entry_points = {'console_scripts':
          ['seekr_download_gencode = seekr.console_scripts:console_download_gencode',
           'seekr_canonical_gencode = seekr.console_scripts:console_canonical_gencode',
           'seekr_kmer_counts = seekr.console_scripts:console_kmer_counts',
           'seekr_pearson = seekr.console_scripts:console_pearson',
           'seekr_visualize_distro = seekr.console_scripts:console_visualize_distro',
           'seekr_norm_vectors = seekr.console_scripts:console_norm_vectors',
           'seekr_gen_rand_rnas = seekr.console_scripts:console_gen_rand_rnas',
           'seekr_graph = seekr.console_scripts:console_graph',
           'seekr_pwm = seekr.console_scripts:console_pwm',
           'seekr_domain_pearson = seekr.console_scripts:console_domain_pearson',
           'seekr = seekr.console_scripts:console_seekr_help']})



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Description
-----------
Generate a matrix of Pearson similarities from two kmer count files.

Examples
--------
Generate a small, plain text .csv file:

```
import pandas as pd
from seekr.pearson import pearson

dist = pearson(counts1=example_2mers,
               counts2=example_2mers)
pd.DataFrame(dist).to_csv('example_vs_example_2mers.csv')

Notes
-----

Issues
------
Any issues can be reported to https://github.com/CalabreseLab/seekr/issues

---
"""

import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from pickle import dump

try:
    from scipy.stats import t
except ImportError:
    pass

from seekr.my_tqdm import my_tqdm
from seekr.fasta_reader import Reader
from seekr.kmer_counts import BasicCounter, Log2
from seekr.utils import get_adj


def pearson(counts1, counts2, row_standardize=True, outfile=None):
    """Calculates a column standardized Pearson correlation matrix"""
    if row_standardize:
        counts1 = (counts1.T - np.mean(counts1, axis=1)).T
        counts1 = (counts1.T / np.std(counts1, axis=1)).T
        counts2 = (counts2.T - np.mean(counts2, axis=1)).T
        counts2 = (counts2.T / np.std(counts2, axis=1)).T

    # Take the inner product and save
    dist = np.inner(counts1, counts2)/counts1.shape[1]
    if outfile:
        np.save(outfile, dist)
    return dist


def visualize_distro(adj, out_path, sample=None):
    adj = get_adj(adj)
    if isinstance(adj, pd.DataFrame):
        adj = adj.values
    if sample is not None:
        if not 0 < sample <= 1:
            raise ValueError('Value of sample must satisfy: 0 < sample <= 1')
        size = int(len(adj) * sample)
        rows = np.random.choice(len(adj), size, replace=False)
        cols = np.random.choice(adj.shape[1], size, replace=False)
        adj = adj[rows][:, cols]
    flat = adj.ravel()
    mean = flat.mean()
    std = flat.std()
    std1 = mean + std
    std2 = mean + (2*std)
    ax = sns.distplot(flat, label=f'Distro (n={flat.size})')
    y_max = ax.get_ylim()[1]
    plt.plot((mean, mean), (0, y_max/2), label='Mean')
    plt.plot((std1, std1), (0, y_max/2), label='Mean + 1 std. dev.')
    plt.plot((std2, std2), (0, y_max/2), label='Mean + 2 std. dev.')
    plt.xlabel('r-value')
    plt.legend()
    plt.savefig(out_path, bbox_inches='tight', dpi=600)
    return mean, std


def pvalue(dist, N):
    """Calculate the p-values (two tailed) of a Pearson correlation matrix

    Parameters
    ----------
    dist : ndarray
        Pearson correlation matrix
    N : int
        Sample size. (The length of each array used to find the R value)

    Returns
    -------
    pvals : ndarray
        p-value for each R value

    Reference
    ---------
    http://stats.stackexchange.com/questions/120199/calculate-p-value-for-the-correlation-coefficient
    http://stackoverflow.com/questions/17559897/python-p-value-from-t-statistic
    """
    pvals = dist / np.sqrt((1-dist**2) / (N -2))
    pvals = t.sf(np.abs(pvals), N - 2)*2
    return pvals


class DomainPearson:
    """Calculate r-values and percentiles over windows of target sequences for a list of queries.

    Parameters
    ----------
    query_path: str (default=None)
        Path to fasta file containing transcripts of interest (e.g. Xist-2kb).
    target_path: str (default=None)
        Path to second fasta file which will be tiled to find domains similar to query transcripts.
    reference_path: str (default=None)
        Path to third fasta file containing sequences to be used for comparison when calculating
        percentile values of the r-values between the query and targets (e.g. mouse transcriptome).
    r_values_path: str (default=None)
        Path to csv file containing pairwise comparisons between queries and target tiles.
    percentiles_path: str (default=None)
        Path to csv file containing the percentile equivalent of each element in r_value_path's csv.
    mean: bool, np.array, str (default=True)
        Set the mean to 0 for each kmer/column of the count matrix.
        If str, provide path to a previously calculated mean array.
        Can be produced by `seekr_norm_vectors`.
    std: bool, np.array, str (default=True)
        Set the std. dev. to 1 for each kmer/column of the count matrix.
        If str, provide path to a previously calculated std array.
        Can be produced by `seekr_norm_vectors`.
    log2: Log2 (default=Log2.post)
        Log2 transformation can occur pre- or post-standardization, or not at all.
    k: int (default=6)
        Size of kmer to be counted.
    window: int (default=1000)
        Size of tile/domain to be created from target transcripts for comparison against queries.
    slide: int (default=100)
        Number of basepairs to move along target transcript before creating another tile/domain.

    Attributes
    ----------
    row_labels: List[str]
        Names of query transcripts from fasta file.
    column_labels List[str]
        Names of tiles as created at "{target fasta header}_{tile number}".
    r_values: DataFrame
        n by m array of Pearson r-values.
        n is the number of query transcripts and m is the total number of target tiles/domains.
    percentiles: DataFrame
        Percentile equivalent of each element in r_values, across each row.

    Notes
    -----
    This code is inefficient.
    Email me (jessime.kirk@gmail.com) if you're running into memory/speed issues.
    There are 'clever' ways of doing this that are a lot messier and harder to maintain.
    """

    def __init__(self, query_path=None, target_path=None, reference_path=None, r_values_path=None,
                 percentiles_path=None, mean=True, std=True, log2=Log2.post,
                 k=6, window=1000, slide=100):
        self.query_path = query_path
        self.target_path = target_path
        if reference_path is None and percentiles_path is not None:
            msg = "To calculate percentiles, pass a path to a reference fasta file."
            raise ValueError(msg)
        self.reference_path = reference_path
        self.r_values_path = r_values_path
        self.percentiles_path = percentiles_path
        self.mean = mean
        self.std = std
        self.log2 = log2
        self.k = k
        self.window = window
        self.slide = slide

        self.row_labels = []
        self.column_labels = []
        self.r_values = None
        self.percentiles = None

    def make_query_counts(self):
        """Create normalized counts for all query transcripts.

        Returns
        -------
        query_counts: ndarray
            Normalized kmer profiles for each query transcript.
        """
        counter = BasicCounter(self.query_path, k=self.k, mean=self.mean, std=self.std,
                               log2=self.log2)
        counter.get_counts()
        return counter.counts

    def get_target_tile_counts(self, target):
        """Create normalized counts for all tiles in a single target transcript.

        If a target sequence is smaller than window size, the whole sequence is used for comparison.

        Parameters
        ----------
        target: str
            Sequence from which to create tiles and kmer counts.

        Returns
        -------
        target_counts: ndarray
            Normalized kmer profiles for each tile in target transcript.
        """
        tiles = []
        upper = max(1, len(target) - self.window + 1)
        for i in range(0, upper, self.slide):
            end = i + self.window
            tiles.append(target[i: end])
        tiles[-1] += target[end:]
        counter = BasicCounter(k=self.k, mean=self.mean, std=self.std, log2=self.log2)
        counter.seqs = tiles
        counter.get_counts()
        return counter.counts

    def update_column_labels(self, target_header, target_counts):
        """Add new column labels from all tiles in latest target transcript.

        Parameters
        ----------
        target_header: str
            Name of target from fasta file.
        target_counts: ndarray
            Normalized kmer profiles for each tile in target transcript.
        """
        self.column_labels.extend(target_header + f'_{i}' for i in range(len(target_counts)))

    def compare_query_target(self, query_counts, target_header, target):
        """Calculare r-values for """
        target_counts = self.get_target_tile_counts(target)
        self.update_column_labels(target_header, target_counts)
        r_vals = pearson(query_counts, target_counts)
        return r_vals

    def r_values2df(self, r_values):
        """Convert r_values from list of numpy arrays to a labeled DataFrame."""
        self.r_values = np.hstack(r_values)
        self.r_values = pd.DataFrame(data=self.r_values,
                                     index=self.row_labels,
                                     columns=self.column_labels)

    def percentileofscore(self, a, score):
        """Same as scipy's percentileofscore where kind='rank'.

        See: https://github.com/scipy/scipy/blob/master/scipy/stats/stats.py#L1846

        Notes
        -----
        This is copied here to avoid another large dependency.
        If scipy eventually becomes necessary, remove this code.
        """
        left = np.count_nonzero(a < score)
        right = np.count_nonzero(a <= score)
        pct = (right + left + (1 if right > left else 0)) * 50.0/len(a)
        return pct

    def calc_percentiles(self, query_counts):
        """Find percentiles of elements in r_values, across each row/query, relative to a reference.

        Notes
        -----
        Repeatedly calling `percentileofscore` is probably really slow.
        There should be some way of speeding things up, similar to `calc_internal_percentiles`.
        I don't have enough time to look into it now.

        Returns
        -------
        percentiles_df: DataFame
            Percentile equivalent of each element in r_values, across each row.
        """
        counter = BasicCounter(infasta=self.reference_path, k=self.k, mean=self.mean, std=self.std,
                               log2=self.log2)
        counter.get_counts()
        query_vs_ref_rvals = pearson(query_counts, counter.counts)[0]
        percentiles_df = pd.DataFrame(columns=self.r_values.columns)
        for index, row in self.r_values.iterrows():
            percentiles = []
            for query_vs_tile_rval in row:
                percentiles.append(self.percentileofscore(query_vs_ref_rvals,
                                                          query_vs_tile_rval))
            percentiles_df.loc[index] = percentiles
        return percentiles_df

    def calc_internal_percentiles(self):
        """Calculate percentile scores of each element in r_values, across each row/query.

        Notes
        -----
        This method doesn't need a reference to calculate percentiles.
        It currently isn't called durning DomainPearson.run()

        Returns
        -------
        percentiles: DataFame
            Percentile equivalent of each element in r_values, across each row.
        """
        percentiles = self.r_values.rank(axis=1, pct=True) * 100
        return percentiles

    def save(self):
        """Save csv files of r_values and percentiles, using limited precision float values."""
        if self.r_values_path is not None:
            self.r_values.to_csv(self.r_values_path, float_format='%.4f')
        if self.percentiles_path is not None:
            self.percentiles.to_csv(self.percentiles_path, float_format='%.4f')

    def run(self):
        self.row_labels = Reader(self.query_path).get_headers()
        query_counts = self.make_query_counts()
        target_reader = Reader(self.target_path)
        target_headers_seqs = target_reader.get_data(tuples_only=True)
        r_values = []
        for target_header, target in my_tqdm()(target_headers_seqs):
            r_values.append(self.compare_query_target(query_counts, target_header, target))
        self.r_values2df(r_values)
        self.percentiles = self.calc_percentiles(query_counts)
        self.save()



import sys
import argparse
import numpy as np
import pandas as pd

from seekr import fasta
from seekr import graph
from seekr import pearson
from seekr.kmer_counts import BasicCounter, Log2

# TODO (Dan) fix names
from seekr.pwm import CountsWeighter

DOWNLOAD_GENCODE_DOC = """
Description
-----------
Download fasta files from https://www.gencodegenes.org/

The one parameter that must be passed is 'biotype'.
Its value must be one of:
* 'all' : Nucleotide sequences of all transcripts on the reference chromosomes
* 'pc' : Nucleotide sequences of coding transcripts on the reference chromosomes
* 'lncRNA' : Nucleotide sequences of long non-coding RNA transcripts on the reference chromosomes

Examples
--------
To download all human transcripts of the latest release into a fasta file:
    $ seekr_download_gencode all

To do the same for mouse:
    $ seekr_download_gencode all -s mouse

To get lncRNAs from the M5 release of mouse:
    $ seekr_download_gencode lncRNA -s mouse -r M5

If you want to leave the fasta file gzipped:
    $ seekr_download_gencode all -z

Issues
------
Any issues can be reported to https://github.com/CalabreseLab/seekr/issues

---
"""

CANONICAL_GENCODE_DOC = """
Description
-----------
Filter GENCODE fasta file for only transcripts ending in 01.

This is based on the common names provided by GENCODE.
No strict guarantees are made about the relationship between genes and transcripts.

Examples
--------
To filter transcripts ending in 01, an input and output fasta file are required:
    $ seekr_canonical_gencode rnas.fa rnas01.fa

If you want to specifically find transcripts with the ending 001:
    $ seekr_canonical_gencode rnas.fa rnas01.fa -z 2

To enforce one isoform per ENSG id (specifically, the smallest 01 isoform):
    $ seekr_canonical_gencode rnas.fa rnas01_1per.fa -u

Issues
------
Any issues can be reported to https://github.com/CalabreseLab/seekr/issues

---
"""

KMER_COUNTS_DOC = """
Description
-----------
Generates a kmer count matrix of m rows by n columns,
where m is the number of transcripts in a fasta file and n is 4^kmer.

Examples
--------
The default settings take a .fa file and produce a labeld csv file:
    $ seekr_kmer_counts rnas.fa -o out.csv

To get a compact and efficient .npy file, set the binary flag:
    $ seekr_kmer_counts rnas.fa -o out.npy -b

You can change also change the size of the kmer you're using, and prevent normalization:
    $ seekr_kmer_counts rnas.fa -o out.csv -k 4 -uc -us -l none

If you ever do not want labels on a csv file:
    $ seekr_kmer_counts rnas.fa -o out.csv -rl

Notes
-----
For more sophisticated options, you cannot use the command line, but need python instead.

To pass --log 1 argument for pre-zscore log-transform of k-mer counts, seekr_norm_vectors MUST be
    run with the -cl flag. This log transforms the reference counts for appropriate mean and std calcs

Issues
------
Any issues can be reported to https://github.com/CalabreseLab/seekr/issues

---
"""

PEARSON_DOC = """
Description
-----------
Generate a matrix of Pearson similarities from two kmer count files.

Examples
--------
The default settings accept two csv files and output a third csv file.
    $ seekr_pearson kc_out.csv kc_out.csv -o out.csv

The only other options besides the `-o` flag control binary versus plain text input and output.
If you have a binary input file (i.e. a .npy file) and also want a binary output file, you can do:
    $ seekr_pearson kc_out.npy kc_out.npy -o out.npy -bi -bo

Notes
-----
For more sophisticated options, you cannot use the command line, but need python instead.

Issues
------
Any issues can be reported to https://github.com/CalabreseLab/seekr/issues

---
"""


VISUALIZE_DISTRO_DOC = """
Description
-----------
Generate an image showing the distribution of all Pearson r-values.
This can be useful for determining a threshold for the adjacency matrix.

Examples
--------
You must pass an adjacency matrix and an output path.
    $ seekr_visualize_distro adj.csv adj.pdf

For large arrays, it's likely sufficient to visualize a portion of the adjacency matrix.
You can pass a float between 0 and 1 to the `-s` flag:
    $ seekr_visualize_distro adj.csv adj.pdf -s .1

Notes
-----
For more sophisticated options, you cannot use the command line, but need python instead.

Issues
------
Any issues can be reported to https://github.com/CalabreseLab/seekr/issues

---
"""


NORM_VECTORS_DOC = """
Description
-----------
Generate two .npy files from a .fa file to use as normalization vectors for other .fa files.

Examples
--------
The default setting accept a single fasta file.
    $ seekr_norm_vectors gencode.fa

If you want to specify paths for the output files, or choose a different kmer size:
    $ seekr_norm_vectors gencode.fa -k 5 -mv mean_5mers.npy -sv std_5mers.npy

If pre-zscore log transform is desired, you must pass the `--log2 pre` flag to log transform
    the reference k-mer counts

Issues
------
Any issues can be reported to https://github.com/CalabreseLab/seekr/issues

---
"""

GEN_RAND_RNAS_DOC = """
Description
-----------
Given a .fa file, create a second .fa file with random RNAs based on original RNAs.

Users control how similar the synthetic RNAs are to the originals.
Gross scale of kmer content can be controlled by setting kmer conservation size.
Fine scale control of similarity can be set by number of SNP mutations.

Examples
--------
The two required positional arguments are an input and output path to fasta files.
This will shuffle the nucleotides for each RNA:
    $ seekr_gen_rand_rnas rnas.fa rnas_rand.fa

To conserve kmer content for k > 1, choose a different kmer size:
    $ seekr_gen_rand_rnas rnas.fa rnas_rand.fa -k 2

Shuffling kmers is random. To reproduce output between runs, set a seed:
    $ seekr_gen_rand_rnas rnas.fa rnas_rand.fa -s 0

It may be useful to conserve the kmer content of the whole fasta file.
Setting the `--group` flag loses conservation of individual sequences,
in preference for producing RNAs with a kmer frequency equal to background frequency.
Using `--group` still produces the same number of output RNAs as input.
Note: this may segfault on large fasta files.
    $ seekr_gen_rand_rnas rnas.fa rnas_rand.fa -k 2 -g

In some cases, it is useful to have more fine-grained control over final kmer content.
Ex: when conserving large kmers, it may be impossible to shuffle shorter seqs.
Ex: if you want to produce a sequence with an exact Pearson's r-value to the original.
A number of random SNP mutations can be made in addition to shuffling.
Use the --mutation flag to set the approximate number of SNP mutations.
Note: Because the new nucleotide may be the same as the old, -m is approximate.
    $ seekr_gen_rand_rnas rnas.fa rnas_rand.fa -m 100

Issues
------
Any issues can be reported to https://github.com/CalabreseLab/seekr/issues

---
"""


GRAPH_DOC = """
Description
-----------
Find communities of transcripts from an adjacency matrix.

Examples
--------
Default setting accept a csv file, and a threshold value.
The csv file should be the product of seekr_pearson, or some other adjacency matrix.
The threshold is the value below which edges are removed from the graph.
seekr_pearson_distro can be run to suggest a value for the threshold.
A gml file contain the graph and communities will be produced.
    $ seekr_graph adj.csv .13 -g graph.gml

For a cleaner csv file of just community information:
    $ seekr_graph adj.csv .5 -g graph.gml -c communities.csv

To change the resolution parameter (gamma) for louvain/leidenalg:
    $ seekr_graph adj.csv .1 -g graph.gml -r 1.5

To change the cap of the number of communities found, and set the seed:
    $ seekr_graph adj.csv .1 -g graph.gml -n 10 -s 0

Numpy files are also valid input:
    $ seekr_graph adj.npy .1 -g graph.gml

Issues
------
Any issues can be reported to https://github.com/CalabreseLab/seekr/issues

---
"""


PWM_DOC = """
Description
-----------
Weight kmer profiles by protein binding PWMs to infer protein binding likelihood.

Examples
--------
A standard run of this tool needs three things:
1. A directory of PWM files
2. A counts file (produced from seekr_kmer_counts)
3. An output path
    $ seekr_pwm path/to/pwms/ kc_out.csv -o pwm_weight_sums.csv

Numpy files can also be passed as input, but .csv files are the only output:
    $ seekr_pwm path/to/pwms/ kc_out.npy -o pwm_weight_sums.csv

The kmer size can also be passed.
It should match the counts file.
Unlike most other seekr tools k=5 is the default for this tool.
    $ seekr_pwm path/to/pwms/ kc_out.npy -k 6 -o pwm_weight_sums.csv

Issues
------
Any issues can be reported to https://github.com/CalabreseLab/seekr/issues

---
"""


DOMAIN_PEARSON_DOC = """
Description
-----------
# Find domains of similarity between query transcripts and tiles of target transcripts.

Examples
--------
This tool requires several pieces of data:
1. A fasta file containing query sequences.
2. A second fasta file containing target sequences which will be tiled.
3. A mean vector for normalization (e.g. from `seekr_norm_vectors`).
4. A std vector for standardization (e.g. from `seekr_norm_vectors`).

For brevity in the documentation below,
we will assume that these required data have been stored in a variable:
    $ REQUIRED="queries.fa targets.fa mean.npy std.npy"

To see the r-values, pass a location for storing them in a csv file.
    $ seekr_domain_pearson $REQUIRED -r r_values.csv

Intepretation of r-value elements can be aided by viewing them as percentiles.
If you want percentiles, you must also pass a reference fasta path:
    $ seekr_domain_pearson $REQUIRED -r r_values.csv -p percentiles.csv -rp reference.fa

Parameters you might pass to `seekr_kmer_counts` can also be passed.
If you change --kmer, ensure that your mean.npy and std.npy files match:
    $ seekr_domain_pearson $REQUIRED -r r_values.csv -l none -k 5

You can also change the size of the domain,
and how far you slide along the target sequence before creating another domain:
    $ seekr_domain_pearson $REQUIRED -r r_values.csv -w 1200 -s 150

Issues
------
Any issues can be reported to https://github.com/CalabreseLab/seekr/issues

---
"""


def _parse_args_or_exit(parser):
    """"""
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(0)
    return parser.parse_args()


def _run_download_gencode(biotype, species, release, out_path, unzip):
    # Note: This function is separated for testing purposes.
    downloader = fasta.Downloader()
    downloader.get_gencode(biotype, species, release, out_path, unzip)


def console_download_gencode():
    assert sys.version_info[0] == 3, 'Python version must be 3.x'
    parser = argparse.ArgumentParser(usage=KMER_COUNTS_DOC,
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('biotype',
                        help=("Name of Genocde set to download. "
                              "Must be one of ('all', 'pc', 'lncRNA')."))
    parser.add_argument('-s', '--species', default='human',
                        help=" Name of species. Must be one of: ('human' or 'mouse').")
    parser.add_argument('-r', '--release', default=None,
                        help=("Name of specific release to download (e.g. 'M5'). "
                              "If None, download latest release."))
    parser.add_argument('-o', '--out_path', default=None,
                        help="Path to location for fasta file. Default will save by release name.")
    parser.add_argument('-z', '--zip', action='store_false',
                        help="Set if you do not want to gunzip fasta file after downloading.")
    args = _parse_args_or_exit(parser)
    _run_download_gencode(args.biotype, args.species, args.release, args.out_path, args.zip)


def _run_canonical_gencode(in_fasta, out_fasta, zeros, unique_per_gene):
    # Note: This function is separated from console_kmer_counts for testing purposes.
    maker = fasta.Maker(in_fasta, out_fasta)
    maker.filter1(zeros, unique_per_gene)


def console_canonical_gencode():
    assert sys.version_info[0] == 3, 'Python version must be 3.x'
    parser = argparse.ArgumentParser(usage=CANONICAL_GENCODE_DOC,
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('in_fasta', help='Full path of fasta file.')
    parser.add_argument('out_fasta', help='Full path of filtered fasta file.')
    parser.add_argument('-z', '--zeros', default=1,
                        help='Number of zeroes needed to be considered canonical.')
    parser.add_argument('-u', '--unique_per_gene', action='store_true',
                        help='Set to enforce a limit of one isoform per ENSG id.')
    args = _parse_args_or_exit(parser)
    _run_canonical_gencode(args.in_fasta, args.out_fasta, args.zeros, args.unique_per_gene)


def _run_kmer_counts(fasta, outfile, kmer, binary, centered, standardized,
                     log2, remove_labels, mean_vector, std_vector, alphabet):
    # Note: This function is separated from console_kmer_counts for testing purposes.
    mean = mean_vector or centered
    std = std_vector or standardized
    label = not remove_labels
    counter = BasicCounter(fasta, outfile, kmer, binary,
                           mean, std, Log2[log2], label=label, alphabet=alphabet)
    counter.make_count_file()


def console_kmer_counts():
    assert sys.version_info[0] == 3, 'Python version must be 3.x'
    parser = argparse.ArgumentParser(usage=KMER_COUNTS_DOC,
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('fasta', help='Full path of fasta file.')
    parser.add_argument('-o', '--outfile', default='counts.seekr',
                        help='Name of file to save counts to.')
    parser.add_argument('-k', '--kmer', default=6,
                        help='Length of kmers you want to count.')
    parser.add_argument('-b', '--binary', action='store_true',
                        help='Set if output should be a .npy file.')
    parser.add_argument('-uc', '--uncentered', action='store_false',
                        help='Set if output should not have the mean subtracted.')
    parser.add_argument('-us', '--unstandardized', action='store_false',
                        help='Set if output should not be divided by the standard deviation.')
    parser.add_argument('-l', '--log2', default=Log2.post.name,
                        choices=[l2.name for l2 in Log2],
                        help='Decided if and when to log transform counts')
    parser.add_argument('-rl', '--remove_labels', action='store_true',
                        help='Set to save without index and column labels.')
    parser.add_argument('-mv', '--mean_vector', default=None,
                        help='Optional path to mean vector numpy file.')
    parser.add_argument('-sv', '--std_vector', default=None,
                        help='Optional path to std vector numpy file.')
    parser.add_argument('-a', '--alphabet', default='AGTC',
                        help='Valid letters to include in kmer.')
    args = _parse_args_or_exit(parser)
    _run_kmer_counts(args.fasta, args.outfile, int(args.kmer), args.binary, args.uncentered,
                     args.unstandardized, args.log2, args.remove_labels, args.mean_vector,
                     args.std_vector, args.alphabet)


def _run_pearson(counts1, counts2, outfile, binary_input, binary_output):
    # Note: This function is separated for testing purposes.
    names1 = None
    names2 = None
    if binary_input:
        counts1 = np.load(counts1)
        counts2 = np.load(counts2)
    else:
        counts1 = pd.read_csv(counts1, index_col=0)
        counts2 = pd.read_csv(counts2, index_col=0)
        names1 = counts1.index.values
        names2 = counts2.index.values

    if binary_output:
        pearson.pearson(counts1, counts2, outfile=outfile)
    else:
        dist = pearson.pearson(counts1, counts2)
        dist = pd.DataFrame(dist, names1, names2)
        dist.to_csv(outfile)


def console_pearson():
    assert sys.version_info[0] == 3, 'Python version must be 3.x'
    parser = argparse.ArgumentParser(usage=PEARSON_DOC,
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('counts1',
                        help='Full path of a count file produced by kmer_counts.py.')
    parser.add_argument('counts2',
                        help=('Full path of a second count file produced by kmer_counts.py. '
                              'This can be the same path as the first counts file.'))
    parser.add_argument('-o', '--outfile', default='pearson.seekr',
                        help='Path of file to save similarities to.')
    parser.add_argument('-bi', '--binary_input', action='store_true',
                        help='Set if the input will be a .npy file.')
    parser.add_argument('-bo', '--binary_output', action='store_true',
                        help='Set if output should be a .npy file.')
    args = _parse_args_or_exit(parser)
    _run_pearson(args.counts1, args.counts2, args.outfile,
                 args.binary_input, args.binary_output)


def _run_visualize_distro(adj, out_path, sample):
    if sample is not None:
        sample = float(sample)
    mean, std = pearson.visualize_distro(adj, out_path, sample)
    print('Mean: ', mean)
    print('Std. Dev.: ', std)


def console_visualize_distro():
    assert sys.version_info[0] == 3, 'Python version must be 3.x'
    parser = argparse.ArgumentParser(usage=VISUALIZE_DISTRO_DOC,
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('adj',
                        help='Path to either .csv or .npy file, representing adjacency matrix')
    parser.add_argument('out_path', help='Full path of a output image.')
    parser.add_argument('-s', '--sample', default=None,
                        help='Float representing random portion of adj to visualize.')
    args = _parse_args_or_exit(parser)
    _run_visualize_distro(args.adj, args.out_path, args.sample)


def _run_norm_vectors(fasta, mean_vector, std_vector, log2, kmer):
    counter = BasicCounter(fasta, k=int(kmer), log2=Log2[log2])
    counter.get_counts()
    np.save(mean_vector, counter.mean)
    np.save(std_vector, counter.std)


def console_norm_vectors():
    assert sys.version_info[0] == 3, 'Python version must be 3.x'
    parser = argparse.ArgumentParser(usage=NORM_VECTORS_DOC,
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('fasta', help='path to .fa file')
    parser.add_argument('-mv', '--mean_vector', default='mean.npy',
                        help='path to output mean vector')
    parser.add_argument('-sv', '--std_vector', default='std.npy',
                        help='path to output standard deviation vector')
    parser.add_argument('-l', '--log2', default=Log2.post.name,
                        choices=[l2.name for l2 in Log2],
                        help='Decided if and when to log transform counts')
    parser.add_argument('-k', '--kmer', default=6,
                        help='length of kmers you want to count')
    args = _parse_args_or_exit(parser)
    _run_norm_vectors(args.fasta, args.mean_vector, args.std_vector, args.log2, int(args.kmer))


def _run_graph(adj, threshold, gml_path, csv_path, louvain, resolution, n_comms, seed):
    # Note: This function is separated for testing purposes.
    leiden = not louvain
    if seed is not None:
        seed = int(seed)
    maker = graph.Maker(adj, gml_path, csv_path, leiden, threshold, resolution, n_comms, seed)
    maker.make_gml_csv_files()


def console_graph():
    assert sys.version_info[0] == 3, 'Python version must be 3.x'
    parser = argparse.ArgumentParser(usage=GRAPH_DOC,
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('adj',
                        help='Path to either .csv or .npy file, representing adjacency matrix')
    parser.add_argument('threshold', help=('Value for thresholding adjacency matrix. '
                                           'Below this limit, all edges are 0.'))
    parser.add_argument('-g', '--gml_path', default=None,
                        help='Path to output graph file in .gml format')
    parser.add_argument('-c', '--csv_path', default=None,
                        help='Path to output community file in .csv format')
    parser.add_argument('-l', '--louvain',  action='store_true',
                        help='If set, use Louvain for community detection instead of Leiden.')
    parser.add_argument('-r', '--resolution', default=1,
                        help=' Resolution parameter for community detection algorithm')
    parser.add_argument('-n', '--n_comms', default=5,
                        help='Number of communities to find. This does not count a null community.')
    parser.add_argument('-s', '--seed', default=None,
                        help='An integer to create reproducible results between runs.')
    args = _parse_args_or_exit(parser)
    _run_graph(args.adj, float(args.threshold), args.gml_path, args.csv_path, args.louvain,
               float(args.resolution), int(args.n_comms), args.seed)


def _run_gen_rand_rnas(in_fasta, out_fasta, kmer, mutations, seed, group):
    # Note: This function is separated for testing purposes.
    # TODO do something with group?
    kmer = int(kmer)
    mutations = int(mutations)
    if seed is not None:
        seed = int(seed)
    individual = not group
    rand_maker = fasta.RandomMaker(in_fasta, out_fasta, kmer, mutations, seed, individual)
    rand_maker.synthesize_random()


def console_gen_rand_rnas():
    assert sys.version_info[0] == 3, 'Python version must be 3.x'
    parser = argparse.ArgumentParser(usage=GEN_RAND_RNAS_DOC,
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('in_fasta', help='path to .fa file')
    parser.add_argument('out_fasta', help='path to new .fa file')
    parser.add_argument('-k', '--kmer', default=1,
                        help='Length of kmers you want to conserve')
    parser.add_argument('-m', '--mutations', default=0,
                        help='Number of SNP mutations to make in RNA')
    parser.add_argument('-s', '--seed', default=None,
                        help='An integer to create reproducible results between runs.')
    parser.add_argument('-g', '--group', action='store_true',
                        help='Set to concatenate RNAs before shuffling and mutating.')
    args = _parse_args_or_exit(parser)
    _run_gen_rand_rnas(args.in_fasta, args.out_fasta, int(args.kmer), int(args.mutations),
                       args.seed, args.group)


def _run_pwms(pwm_dir, counts, kmer, out_path):
    # TODO (Dan) name this function
    # Note: This function is separated for testing purposes.
    counts_weighter = CountsWeighter(pwm_dir, counts, kmer, out_path)
    counts_weighter.run()


def console_pwm():
    # TODO (Dan) name this function
    assert sys.version_info[0] == 3, 'Python version must be 3.x'
    parser = argparse.ArgumentParser(usage=PWM_DOC,
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('pwm_dir', help='Path to directory containing PWM files.')
    parser.add_argument('counts', help='Path to kmer_counts file.')
    parser.add_argument('-k', '--kmer', default=5,
                        help='Length of kmer.')
    parser.add_argument('-o', '--out_path',
                        help='Path to new csv file containing weighted count sums.')
    args = _parse_args_or_exit(parser)
    # TODO (Dan) update name
    _run_pwms(args.pwm_dir, args.counts, int(args.kmer), args.out_path)


def _run_domain_pearson(query_path, target_path, reference_path, mean, std, r_values,
                        percentiles, kmer, log2, window, slide):
    # Note: This function is separated for testing
    domain_pearson = pearson.DomainPearson(query_path, target_path, reference_path, r_values,
                                           percentiles, mean, std, Log2[log2], kmer, window, slide)
    domain_pearson.run()


def console_domain_pearson():
    assert sys.version_info[0] == 3, 'Python version must be 3.x'
    parser = argparse.ArgumentParser(usage=DOMAIN_PEARSON_DOC,
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('query_path',
                        help='Path to fa file containing transcripts of interest (e.g. Xist-2kb).')
    parser.add_argument('target_path', help=('Path to second fa file which will be tiled to find '
                                             'domains similar to query transcripts.'))
    parser.add_argument('mean', help='Path to npy file containing mean array for normalization.')
    parser.add_argument('std', help='Path to npy file containing std array for standardization.')
    parser.add_argument('-rp', '--reference_path', default=None,
                        help=('Path to third fasta file containing sequences to be used for '
                              'comparison when calculating percentile values of the r-values '
                              'between the query and targets (e.g. mouse transcriptome).'))
    parser.add_argument('-r', '--r_values', help='Path to new csv file for storing r-values.')
    parser.add_argument('-p', '--percentiles', help='Path to new csv file for storing percentiles.')
    parser.add_argument('-k', '--kmer', default=6,
                        help='Length of kmers you want to count.')
    parser.add_argument('-l', '--log2', default=Log2.post.name,
                        choices=[l2.name for l2 in Log2],
                        help='Decided if and when to log transform counts')
    parser.add_argument('-w', '--window', default=1000,
                        help=('Size of tile/domain to be created from target transcripts for '
                              'comparison against queries.'))
    parser.add_argument('-s', '--slide', default=100,
                        help=('Number of basepairs to move along target transcript before creating '
                              'another tile/domain.'))
    args = _parse_args_or_exit(parser)
    _run_domain_pearson(args.query_path, args.target_path, args.reference_path, args.mean, args.std,
                        args.r_values, args.percentiles, int(args.kmer), args.log2,
                        int(args.window), args.slide)


def console_seekr_help():
    intro = ('Welcome to SEEKR! \n'
             'Below is a description of all SEEKR commands.\n'
             'For additional help see the README at: \n'
             'https://github.com/CalabreseLab/seekr.\n\n')
    print(intro)
    cmds2doc = {'seekr_download_gencode': DOWNLOAD_GENCODE_DOC,
                'seekr_canonical_gencode': CANONICAL_GENCODE_DOC,
                'seekr_norm_vectors': NORM_VECTORS_DOC,
                'seekr_kmer_counts': KMER_COUNTS_DOC,
                'seekr_pearson': PEARSON_DOC,
                'seekr_visualize_distro': VISUALIZE_DISTRO_DOC,
                'seekr_graph': GRAPH_DOC,
                'seekr_gen_rand_rnas': GEN_RAND_RNAS_DOC,
                'seekr_pmw': PWM_DOC,
                'seekr_domain_pearson': DOMAIN_PEARSON_DOC}
    for c, d in cmds2doc.items():
        print(f"{'='*25}\n{c}\n{'='*25}\n{d}")
    conclusion = ('To see a full description of flags and defaults, '
                  'run any of the commands listed above, without any parameters '
                  '(e.g. "$ seekr_graph").')
    print(conclusion)



import numpy as np
import pkg_resources
from seekr import kmer_counts


class TestBasicCounter:

    def _create_basic_counter_with_data(self, **kwargs):
        infasta = 'tests/data/example.fa'
        infasta = pkg_resources.resource_filename('seekr', infasta)
        counter = kmer_counts.BasicCounter(infasta=infasta,
                                           silent=True,
                                           log2=kmer_counts.Log2.post,
                                           **kwargs)
        return counter

    def test_counter_init(self):
        counter = self._create_basic_counter_with_data()
        assert len(counter.seqs) == 5
        assert counter.seqs[0] == 'AAAAAA'

    def test_occurrences_k1(self):
        counter = self._create_basic_counter_with_data(k=1)
        row = np.zeros(4)
        expected = row.copy()
        expected[0] = 1000
        row = counter.occurrences(row, counter.seqs[0])
        assert np.allclose(row, expected)

        row = np.zeros(4)
        expected=row.copy()
        expected[1] = 500
        expected[2] = 500
        row = counter.occurrences(row, counter.seqs[1])
        assert np.allclose(row, expected)

    def test_occurrences_k2(self):
        counter = self._create_basic_counter_with_data(k=2)
        row = np.zeros(16)
        expected=row.copy()

        expected[5] = 454.545
        expected[9] = 90.909
        expected[10] = 454.545
        row = counter.occurrences(row, counter.seqs[1])
        assert np.allclose(row, expected)

    def test_center_true(self):
        counter = self._create_basic_counter_with_data(k=1)
        counts = np.array([[1,2,3,4], [1, -2, 5, 10]], dtype=np.float32)
        counter.counts = counts
        expected = np.array([[0, 2, -1, -3], [0, -2, 1, 3]], dtype=np.float32)
        counter.center()
        assert np.allclose(counter.counts, expected)

    def test_center_vector(self):
        counter = self._create_basic_counter_with_data(k=1)
        counts = np.array([[1, 2, 3, 4], [1, -2, 5, 10]], dtype=np.float32)
        counter.counts = counts
        mean = np.ones(4)
        mean[3] = -1
        counter.mean = mean
        expected = np.array([[0, 1, 2, 5], [0, -3, 4, 11]], dtype=np.float32)
        counter.center()
        assert np.allclose(counter.counts, expected)

    def test_standardize_true(self):
        counter = self._create_basic_counter_with_data(k=1)
        counts = np.array([[1, 2, 3, 4], [0, -2, 5, 10]], dtype=np.float32)
        counter.counts = counts
        expected = np.array([[2, 1, 3, 4/3], [0, -1, 5, 10/3]], dtype=np.float32)
        counter.standardize()
        assert np.allclose(counter.counts, expected)

    def test_standardize_vector(self):
        counter = self._create_basic_counter_with_data(k=1)
        counts = np.array([[1, 2, 3, 4], [0, -2, 5, 10]], dtype=np.float32)
        counter.counts = counts
        std = np.arange(1, 5)
        counter.std = std
        expected = np.array([[1, 1, 1, 1], [0, -1, 5/3, 2.5]], dtype=np.float32)
        counter.standardize()
        assert np.allclose(counter.counts, expected)

    def test_log2_norm(self):
        counter = self._create_basic_counter_with_data(k=1)
        counts = np.array([[1, 2, 3, 4], [0, -2, 5, 10]], dtype=np.float32)
        counts+= np.abs(np.min(counts))
        counter.counts = counts
        counter.log2_norm()
        expected = np.array([[1, 2, 3, 4], [0, -2, 5, 10]], dtype=np.float32)
        expected+=np.abs(np.min(expected))
        expected = np.log2(expected+1)
        assert np.allclose(counter.counts, expected)

    def test_get_counts(self):
        counter = self._create_basic_counter_with_data(k=1)
        counter.get_counts()
        expected = np.array([[2.1798673 , 0.27807194, 0.        , 0.5133058 ],
                             [0.6370419 , 2.1100981 , 2.048016  , 0.5133058 ],
                             [1.2010899 , 1.4672222 , 1.3604679 , 1.8107259 ],
                             [1.2073011 , 1.3895708 , 1.3721647 , 1.8666755 ],
                             [1.318994  , 1.1856667 , 1.5349197 , 1.6688585 ]],
                             dtype=np.float32)

        assert np.allclose(counter.counts, expected, rtol=.0001, atol=.00001)

    def test_get_counts_raw(self):
        counter = self._create_basic_counter_with_data(
            k=2, mean=False, std=False)
        counter.get_counts()
        expected = np.zeros((5, 16))
        expected[0] = counter.occurrences(counter.counts[0], counter.seqs[0])
        expected[1] = counter.occurrences(counter.counts[1], counter.seqs[1])
        expected[2] = counter.occurrences(counter.counts[2], counter.seqs[2])
        expected[3] = counter.occurrences(counter.counts[3], counter.seqs[3])
        expected[4] = counter.occurrences(counter.counts[4], counter.seqs[4])
        assert np.allclose(counter.counts, expected)


__title__ = 'seekr'
__description__ = 'Count small kmer frequencies in nucleotide sequences.'
__url__ = 'https://github.com/CalabreseLab/seekr'
__version__ = '1.5.1'
__author__ = 'Jessime Kirk'
__author_email__ = 'jessime.kirk@gmail.com'
__license__ = 'MIT'
